""" test StoragElement
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import tempfile
import mock
import unittest
import itertools

from diraccfg import CFG

from DIRAC import S_OK
from DIRAC.Resources.Storage.StorageElement import StorageElementItem
from DIRAC.Resources.Storage.StorageBase import StorageBase


from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.private.ConfigurationClient import ConfigurationClient


class fake_SRM2Plugin(StorageBase):
    """Fake SRM2 plugin.
    Only implements the two methods needed
    for transfer, so we can test that it is really this plugin
    that returned
    """

    def putFile(self, lfns, sourceSize=0):
        return S_OK({"Successful": dict.fromkeys(lfns, "srm:putFile"), "Failed": {}})

    def getTransportURL(self, path, protocols=False):
        return S_OK({"Successful": dict.fromkeys(path, "srm:getTransportURL"), "Failed": {}})


class fake_XROOTPlugin(StorageBase):
    """Fake XROOT plugin.
    Only implements the two methods needed
    for transfer, so we can test that it is really this plugin
    that returned
    """

    def putFile(self, lfns, sourceSize=0):
        return S_OK({"Successful": dict.fromkeys(lfns, "root:putFile"), "Failed": {}})

    def getTransportURL(self, path, protocols=False):
        return S_OK({"Successful": dict.fromkeys(path, "root:getTransportURL"), "Failed": {}})


class fake_GSIFTPPlugin(StorageBase):
    """Fake GSIFTP plugin.
    Only implements the two methods needed
    for transfer, so we can test that it is really this plugin
    that returned
    """

    def putFile(self, lfns, sourceSize=0):
        return S_OK({"Successful": dict.fromkeys(lfns, "gsiftp:putFile"), "Failed": {}})

    def getTransportURL(self, path, protocols=False):
        return S_OK({"Successful": dict.fromkeys(path, "gsiftp:getTransportURL"), "Failed": {}})


def mock_StorageFactory_generateStorageObject(storageName, pluginName, parameters, hideExceptions=False):
    """Generate fake storage object"""
    storageObj = StorageBase(storageName, parameters)

    if pluginName == "SRM2":
        storageObj = fake_SRM2Plugin(storageName, parameters)
        storageObj.protocolParameters["InputProtocols"] = ["file", "root", "srm"]
        storageObj.protocolParameters["OutputProtocols"] = ["file", "root", "dcap", "gsidcap", "rfio", "srm"]
    elif pluginName == "File":
        # Not needed to do anything, StorageBase should do it :)
        pass
    elif pluginName == "XROOT":
        storageObj = fake_XROOTPlugin(storageName, parameters)
        storageObj.protocolParameters["InputProtocols"] = ["file", "root"]
        storageObj.protocolParameters["OutputProtocols"] = ["root"]
    elif pluginName == "GSIFTP":
        storageObj = fake_GSIFTPPlugin(storageName, parameters)
        storageObj.protocolParameters["InputProtocols"] = ["file", "gsiftp"]
        storageObj.protocolParameters["OutputProtocols"] = ["gsiftp"]

    storageObj.pluginName = pluginName

    return S_OK(storageObj)


class TestBase(unittest.TestCase):
    """Base test class. Defines all the method to test"""

    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
        side_effect=mock_StorageFactory_generateStorageObject,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def setUp(self, _mk_generateStorage, _mk_isLocalSE, _mk_addAccountingOperation):

        # Creating test configuration file
        self.testCfgFileName = os.path.join(tempfile.gettempdir(), "test_StorageElement.cfg")
        cfgContent = """
    DIRAC
    {
      Setup=TestSetup
    }
    Resources{
      StorageElements{
        StorageA
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocol.0
          {
            Host =
            PluginName = File
            Protocol = file
            Path =
          }
        }
        StorageB
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocol.0
          {
            Host =
            PluginName = SRM2
            Protocol = srm
            Path =
          }
        }
        StorageC
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocol.0
          {
            Host =
            PluginName = XROOT
            Protocol = root
            Path =
          }
        }
        StorageD
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocol.0
          {
            Host =
            PluginName = SRM2
            Protocol = srm
            Path =
          }
          AccessProtocol.1
          {
            Host =
            PluginName = XROOT
            Protocol = root
            Path =
          }
        }
        StorageE
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          WriteProtocols = root
          WriteProtocols += srm
          AccessProtocol.0
          {
            Host =
            PluginName = SRM2
            Protocol = srm
            Path =
          }
          AccessProtocol.1
          {
            Host =
            PluginName = XROOT
            Protocol = root
            Path =
          }
        }
        StorageX
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          WriteProtocols = gsiftp
          AccessProtocols = root
          AccessProtocol.0
          {
            Host =
            PluginName = GSIFTP
            Protocol = gsiftp
            Path =
          }
          AccessProtocol.1
          {
            Host =
            PluginName = XROOT
            Protocol = root
            Path =
          }
        }
        StorageY
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocols = gsiftp
          AccessProtocols += srm
          AccessProtocol.0
          {
            Host =
            PluginName = GSIFTP
            Protocol = gsiftp
            Path =
          }
          AccessProtocol.1
          {
            Host =
            PluginName = SRM2
            Protocol = srm
            Path =
          }
        }
        StorageZ
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          AccessProtocols = root
          AccessProtocols += srm
          WriteProtocols = root
          WriteProtocols += srm
          AccessProtocol.0
          {
            Host =
            PluginName = ROOT
            Protocol = root
            Path =
          }
          AccessProtocol.1
          {
            Host =
            PluginName = SRM2
            Protocol = srm
            Path =
          }
        }
      }

    }
    Operations{
      Defaults
      {
        DataManagement{
          AccessProtocols = fakeProto
          AccessProtocols += root
          WriteProtocols = srm
        }
      }
    }
    """

        with open(self.testCfgFileName, "w") as f:
            f.write(cfgContent)

        # SUPER UGLY: one must recreate the CFG objects of gConfigurationData
        # not to conflict with other tests that might be using a local dirac.cfg
        gConfigurationData.localCFG = CFG()
        gConfigurationData.remoteCFG = CFG()
        gConfigurationData.mergedCFG = CFG()
        gConfigurationData.generateNewVersion()

        gConfig = ConfigurationClient(
            fileToLoadList=[self.testCfgFileName]
        )  # we replace the configuration by our own one.

        self.seA = StorageElementItem("StorageA")
        self.seA.vo = "lhcb"
        self.seB = StorageElementItem("StorageB")
        self.seB.vo = "lhcb"
        self.seC = StorageElementItem("StorageC")
        self.seC.vo = "lhcb"
        self.seD = StorageElementItem("StorageD")
        self.seD.vo = "lhcb"
        self.seE = StorageElementItem("StorageE")
        self.seE.vo = "lhcb"

        self.seX = StorageElementItem("StorageX")
        self.seX.vo = "lhcb"
        self.seY = StorageElementItem("StorageY")
        self.seY.vo = "lhcb"
        self.seZ = StorageElementItem("StorageZ")
        self.seZ.vo = "lhcb"

    def tearDown(self):
        try:
            os.remove(self.testCfgFileName)
        except OSError:
            pass
        # SUPER UGLY: one must recreate the CFG objects of gConfigurationData
        # not to conflict with other tests that might be using a local dirac.cfg
        gConfigurationData.localCFG = CFG()
        gConfigurationData.remoteCFG = CFG()
        gConfigurationData.mergedCFG = CFG()
        gConfigurationData.generateNewVersion()

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_01_negociateProtocolWithOtherSE(self, mk_isLocalSE, mk_addAccounting):
        """Testing negotiation algorithm"""

        # Find common protocol between SRM2 and File
        res = self.seA.negociateProtocolWithOtherSE(self.seB)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], ["file"])

        # Find common protocol between File and SRM@
        res = self.seB.negociateProtocolWithOtherSE(self.seA)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], ["file"])

        # Find common protocol between XROOT and File
        # Nothing goes from xroot to file
        res = self.seA.negociateProtocolWithOtherSE(self.seC)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], [])

        # Find common protocol between File and XROOT
        res = self.seC.negociateProtocolWithOtherSE(self.seA)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], ["file"])

        # Find common protocol between File and File
        res = self.seA.negociateProtocolWithOtherSE(self.seA)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], ["file"])

        # Find common protocol between SRM and SRM
        res = self.seB.negociateProtocolWithOtherSE(self.seB)
        self.assertTrue(res["OK"], res)
        self.assertEqual(sorted(res["Value"]), sorted(["file", "root", "srm"]))

        # Find common protocol between SRM and XROOT
        res = self.seC.negociateProtocolWithOtherSE(self.seB)
        self.assertTrue(res["OK"], res)
        self.assertEqual(sorted(res["Value"]), sorted(["root", "file"]))

        # Find common protocol between XROOT and SRM
        res = self.seC.negociateProtocolWithOtherSE(self.seB)
        self.assertTrue(res["OK"], res)
        self.assertEqual(sorted(res["Value"]), sorted(["root", "file"]))

        # Testing restrictions
        res = self.seC.negociateProtocolWithOtherSE(self.seB, protocols=["file"])
        self.assertTrue(res["OK"], res)
        self.assertEqual(sorted(res["Value"]), ["file"])

        res = self.seC.negociateProtocolWithOtherSE(self.seB, protocols=["nonexisting"])
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], [])

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_02_followOrder(self, _mk_isLocalSE, _mk_addAccounting):
        """Testing If the order of preferred protocols is respected"""

        for permutation in itertools.permutations(["srm", "file", "root", "nonexisting"]):
            permuList = list(permutation)
            # Don't get tricked ! remove cannot be put
            # after the conversion, because it is inplace modification
            permuList.remove("nonexisting")
            res = self.seD.negociateProtocolWithOtherSE(self.seD, protocols=permutation)
            self.assertTrue(res["OK"], res)
            self.assertEqual(res["Value"], permuList)

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_03_multiProtocolThirdParty(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case for storages with several protocols

        Here comes the fun :-)
        Suppose we have endpoints that we can read in root, but cannot write
        If we have root in the accessProtocols and thirdPartyProtocols lists
        but not in the writeProtocols, we should get a root url to read,
        and write with SRM

        We reproduce here the behavior of DataManager.replicate

        """

        thirdPartyProtocols = ["root", "srm"]

        lfn = "/lhcb/fake/lfn"
        res = self.seD.negociateProtocolWithOtherSE(self.seD, protocols=thirdPartyProtocols)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], thirdPartyProtocols)

        # Only the XROOT plugin here implements the geTransportURL
        # that returns what we want, so we know that
        # if the return is successful, it is because of the XROOT
        res = self.seD.getURL(lfn, protocol=res["Value"])
        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)

        srcUrl = res["Value"]["Successful"][lfn]
        self.assertEqual(srcUrl, "root:getTransportURL")

        # Only the SRM2 plugin here implements the putFile method
        # so if we get a success here, it means that we used the SRM plugin
        res = self.seD.replicateFile({lfn: srcUrl}, sourceSize=123, inputProtocol="root")

        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)
        self.assertEqual(res["Value"]["Successful"][lfn], "srm:putFile")

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_04_thirdPartyLocalWrite(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case for storages with several protocols

        Here, we locally define the write protocol to be root and srm
        So we should be able to do everything with XROOT plugin

        """

        thirdPartyProtocols = ["root", "srm"]

        lfn = "/lhcb/fake/lfn"
        res = self.seE.negociateProtocolWithOtherSE(self.seE, protocols=thirdPartyProtocols)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], thirdPartyProtocols)

        res = self.seE.getURL(lfn, protocol=res["Value"])
        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)

        srcUrl = res["Value"]["Successful"][lfn]
        self.assertEqual(srcUrl, "root:getTransportURL")

        res = self.seE.replicateFile({lfn: srcUrl}, sourceSize=123, inputProtocol="root")

        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)
        self.assertEqual(res["Value"]["Successful"][lfn], "root:putFile")

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_05_thirdPartyMix(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case for storages with several protocols

        Here, we locally define the write protocol for the destination, so it should
        all go directly through the XROOT plugin

        """

        thirdPartyProtocols = ["root", "srm"]

        lfn = "/lhcb/fake/lfn"
        res = self.seE.negociateProtocolWithOtherSE(self.seD, protocols=thirdPartyProtocols)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], thirdPartyProtocols)

        res = self.seD.getURL(lfn, protocol=res["Value"])
        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)

        srcUrl = res["Value"]["Successful"][lfn]
        self.assertEqual(srcUrl, "root:getTransportURL")

        res = self.seE.replicateFile({lfn: srcUrl}, sourceSize=123, inputProtocol="root")

        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)
        self.assertEqual(res["Value"]["Successful"][lfn], "root:putFile")

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_06_thirdPartyMixOpposite(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case for storages with several protocols

        Here, we locally define the write protocol for the source, so it should
        get the source directly using XROOT, and perform the put using SRM

        """

        thirdPartyProtocols = ["root", "srm"]

        lfn = "/lhcb/fake/lfn"
        res = self.seD.negociateProtocolWithOtherSE(self.seE, protocols=thirdPartyProtocols)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], thirdPartyProtocols)

        res = self.seE.getURL(lfn, protocol=res["Value"])
        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)

        srcUrl = res["Value"]["Successful"][lfn]
        self.assertEqual(srcUrl, "root:getTransportURL")

        res = self.seD.replicateFile({lfn: srcUrl}, sourceSize=123, inputProtocol="root")

        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)
        self.assertEqual(res["Value"]["Successful"][lfn], "srm:putFile")

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_07_multiProtocolSrmOnly(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case for storages with several protocols

        Here comes the fun :-)
        Suppose we have endpoints that we can read in root, but cannot write
        If we have root in the accessProtocols and thirdPartyProtocols lists
        but not in the writeProtocols, we should get a root url to read,
        and write with SRM

        We reproduce here the behavior of DataManager.replicate

        """

        thirdPartyProtocols = ["srm"]

        lfn = "/lhcb/fake/lfn"
        res = self.seD.negociateProtocolWithOtherSE(self.seD, protocols=thirdPartyProtocols)
        self.assertTrue(res["OK"], res)
        self.assertEqual(res["Value"], thirdPartyProtocols)

        res = self.seD.getURL(lfn, protocol=res["Value"])
        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)

        srcUrl = res["Value"]["Successful"][lfn]
        self.assertEqual(srcUrl, "srm:getTransportURL")

        res = self.seD.replicateFile({lfn: srcUrl}, sourceSize=123, inputProtocol="srm")

        self.assertTrue(res["OK"], res)
        self.assertTrue(lfn in res["Value"]["Successful"], res)
        self.assertEqual(res["Value"]["Successful"][lfn], "srm:putFile")

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_08_multiProtocolFTS(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case FTS replication between storages with several protocols

        Here comes the fun :-)
        Suppose we have endpoints that we can read in root, but cannot write
        If we have root in the accessProtocols and thirdPartyProtocols lists
        but not in the writeProtocols, we should get a root url to read,
        and write with SRM.
        And We should get the proper url for source and destination

        Storage X, Y and Z represents the situation we could now have in LHCb:
          * X is RAL Echo: you read with root, write with gsiftp
          * Y is Gridka: you have gsiftp available for read only
          * Z is CERN EOS: you can do everything with EOS

        This makes it necessary to add gsiftp as third party option to write to ECHO

        """
        thirdPartyProtocols = ["root", "gsiftp", "srm"]
        rankedProtocols = ["root", "gsiftp", "gsidcap", "dcap", "file", "srm", "rfio"]

        lfn = "/lhcb/fake/lfn"

        # RAL -> GRIDKA
        # We should read using root and write through srm
        res = self.seY.generateTransferURLsBetweenSEs(lfn, self.seX, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("root:%s" % lfn, "srm:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("root", "srm"))

        # RAL -> CERN
        # We should read using root and write directly with it
        res = self.seZ.generateTransferURLsBetweenSEs(lfn, self.seX, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("root:%s" % lfn, "root:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("root", "root"))

        # GRIDKA -> RAL
        # We should read using gsiftp and write directly with it
        res = self.seX.generateTransferURLsBetweenSEs(lfn, self.seY, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("gsiftp:%s" % lfn, "gsiftp:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("gsiftp", "gsiftp"))

        # GRIDKA -> CERN
        # We should read using srm and write with root
        res = self.seZ.generateTransferURLsBetweenSEs(lfn, self.seY, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("srm:%s" % lfn, "root:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("srm", "root"))

        # CERN -> RAL
        # We should read using srm and write with gsiftp
        res = self.seX.generateTransferURLsBetweenSEs(lfn, self.seZ, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("srm:%s" % lfn, "gsiftp:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("srm", "gsiftp"))

        # CERN -> GRIDKA
        # We should read using root and write directly with srm
        res = self.seY.generateTransferURLsBetweenSEs(lfn, self.seZ, protocols=rankedProtocols)
        self.assertTrue(res["OK"], res)
        urlPair = res["Value"]["Successful"].get(lfn)
        self.assertTupleEqual(urlPair, ("root:%s" % lfn, "srm:%s" % lfn))
        protoPair = res["Value"]["Protocols"]
        self.assertTupleEqual(protoPair, ("root", "srm"))


class TestSameSE(unittest.TestCase):
    """Tests to compare two SEs together."""

    @mock.patch(
        "DIRAC.Resources.Storage.StorageFactory.StorageFactory._StorageFactory__generateStorageObject",
        side_effect=mock_StorageFactory_generateStorageObject,
    )
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def setUp(self, _mk_generateStorage, _mk_isLocalSE, _mk_addAccountingOperation):

        # Creating test configuration file
        self.testCfgFileName = os.path.join(tempfile.gettempdir(), "test_StorageElement.cfg")
        cfgContent = """
    DIRAC
    {
      Setup=TestSetup
    }
    Resources{
      StorageElements{
        DiskStorageA
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          GFAL2_SRM2
          {
            Host = srm-diskandtape.cern.ch
            SpaceToken = Disk
            Protocol = srm
            Path = /base/pathDisk
          }
        }
        # Same end point as DiskStorageA, but with a different space token
        # So they should be considered the same
        TapeStorageA
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          GFAL2_SRM2
          {
            Host = srm-diskandtape.cern.ch
            Protocol = srm
            SpaceToken = Tape
            Path = /base/pathDisk
          }
        }
        # Normally does not happen in practice, but this is the same as DiskStorageA with more plugins
        DiskStorageAWithMoreProtocol
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          GFAL2_SRM2
          {
            Host = srm-diskandtape.cern.ch
            SpaceToken = Disk
            Protocol = srm
            Path = /base/pathDisk
          }
          GFAL2_GSIFTP
          {
            Host = gsiftp-diskandtape.cern.ch
            SpaceToken = Disk
            Protocol = gsiftp
            Path = /base/pathDisk
          }
        }
        # A different storage
        StorageB
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          GFAL2_GSIFTP
          {
            Host = otherstorage.cern.ch
            SpaceToken = Disk
            Protocol = gsiftp
            Path = /base/pathDisk
          }
        }
        # The same endpoint as StorageB but with differetn base path, so not the same
        StorageBWithOtherBasePath
        {
          BackendType = local
          ReadAccess = Active
          WriteAccess = Active
          GFAL2_GSIFTP
          {
            Host = otherstorage.cern.ch
            SpaceToken = Disk
            Protocol = gsiftp
            Path = /base/otherPath
          }
        }
      }
    }
    Operations{
      Defaults
      {
        DataManagement{
          AccessProtocols = fakeProto
          AccessProtocols += root
          WriteProtocols = srm
        }
      }
    }
    """

        with open(self.testCfgFileName, "w") as f:
            f.write(cfgContent)

        # SUPER UGLY: one must recreate the CFG objects of gConfigurationData
        # not to conflict with other tests that might be using a local dirac.cfg
        gConfigurationData.localCFG = CFG()
        gConfigurationData.remoteCFG = CFG()
        gConfigurationData.mergedCFG = CFG()
        gConfigurationData.generateNewVersion()

        gConfig = ConfigurationClient(
            fileToLoadList=[self.testCfgFileName]
        )  # we replace the configuration by our own one.

        self.diskStorageA = StorageElementItem("DiskStorageA")
        self.diskStorageA.vo = "lhcb"
        self.tapeStorageA = StorageElementItem("TapeStorageA")
        self.tapeStorageA.vo = "lhcb"
        self.diskStorageAWithMoreProtocol = StorageElementItem("DiskStorageAWithMoreProtocol")
        self.diskStorageAWithMoreProtocol.vo = "lhcb"
        self.storageB = StorageElementItem("StorageB")
        self.storageB.vo = "lhcb"
        self.storageBWithOtherBasePath = StorageElementItem("StorageBWithOtherBasePath")
        self.storageBWithOtherBasePath.vo = "lhcb"

    def tearDown(self):
        try:
            os.remove(self.testCfgFileName)
        except OSError:
            pass
        # SUPER UGLY: one must recreate the CFG objects of gConfigurationData
        # not to conflict with other tests that might be using a local dirac.cfg
        gConfigurationData.localCFG = CFG()
        gConfigurationData.remoteCFG = CFG()
        gConfigurationData.mergedCFG = CFG()
        gConfigurationData.generateNewVersion()

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_01_compareSEWithItself(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test case to compare SE together
        """

        for se in (
            self.diskStorageA,
            self.tapeStorageA,
            self.diskStorageAWithMoreProtocol,
            self.storageB,
            self.storageBWithOtherBasePath,
        ):

            self.assertTrue(se.isSameSE(se))

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_02_compareSEThatShouldBeTheSame(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test SEs that should be the considered same
        """

        matchingCouples = (
            (self.diskStorageA, self.diskStorageAWithMoreProtocol),
            (self.diskStorageA, self.tapeStorageA),
        )

        for se1, se2 in matchingCouples:
            self.assertTrue(se1.isSameSE(se2))

    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem._StorageElementItem__isLocalSE",
        return_value=S_OK(True),
    )  # Pretend it's local
    @mock.patch(
        "DIRAC.Resources.Storage.StorageElement.StorageElementItem.addAccountingOperation", return_value=None
    )  # Don't send accounting
    def test_02_compareSEThatShouldBeDifferent(self, _mk_isLocalSE, _mk_addAccounting):
        """
        Test SEs that should be the considered same
        """

        notMatchingCouples = (
            (self.diskStorageA, self.storageB),
            (self.tapeStorageA, self.storageB),
            (self.storageB, self.storageBWithOtherBasePath),
        )

        for se1, se2 in notMatchingCouples:
            self.assertFalse(se1.isSameSE(se2))


if __name__ == "__main__":
    from DIRAC import gLogger

    gLogger.setLevel("DEBUG")
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestBase)

    unittest.TextTestRunner(verbosity=2).run(suite)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import unittest

from mock import MagicMock

from DIRAC import gLogger
from DIRAC.ResourceStatusSystem.PolicySystem.PEP import PEP

# from DIRAC.ResourceStatusSystem.PolicySystem.PDP          import PDP
# from DIRAC.ResourceStatusSystem.PolicySystem.PolicyCaller import PolicyCaller

#############################################################################


class PolicySystemTestCase(unittest.TestCase):
    """Base class for the PDP - PEP test cases"""

    #############################################################################

    def setUp(self):
        gLogger.setLevel("DEBUG")

        self.RSMock = MagicMock()
        self.RMMock = MagicMock()
        self.RMMock.selectStatusElement.return_value = {"OK": True, "Value": "bla"}
        self.mockPDP = MagicMock()


#############################################################################


class PEPSuccess(PolicySystemTestCase):

    #############################################################################

    def test_enforce(self):

        pep = PEP(
            {"ResourceStatusClient": self.RSMock, "ResourceManagementClient": self.RMMock, "SiteStatus": self.RMMock}
        )
        pep.pdp = self.mockPDP
        res = pep.enforce(None)
        self.assertTrue(res["OK"])

        decisionParams = {}
        res = pep.enforce(decisionParams)
        self.assertTrue(res["OK"])

        decisionParams = {"element": "Site", "name": "Site1"}
        decParamsPDP = dict(decisionParams)
        decParamsPDP["active"] = "active"
        self.mockPDP.takeDecision.return_value = {
            "OK": True,
            "Value": {
                "policyCombinedResult": {
                    "PolicyType": ["", ""],
                    "PolicyAction": [("aa", "bb")],
                    "Status": "S",
                    "Reason": "testReason",
                },
                "singlePolicyResults": [
                    {"Status": "Active", "PolicyName": "SAM_CE_Policy", "Reason": "SAM:ok"},
                    {
                        "Status": "Banned",
                        "PolicyName": "DT_Policy_Scheduled",
                        "Reason": "DT:OUTAGE in 1 hours",
                        "EndDate": "2010-02-16 15:00:00",
                    },
                ],
                "decisionParams": decParamsPDP,
            },
        }
        res = pep.enforce(decisionParams)
        self.assertTrue(res["OK"])

        decisionParams = {"element": "Resource", "name": "StorageElement", "statusType": "ReadAccess"}
        res = pep.enforce(decisionParams)
        self.assertTrue(res["OK"])


# class PDPSuccess(PolicySystemTestCase):
#
# #############################################################################
#
#
#   def test_takeDecision(self):
#
#     for granularity in ValidRes:
#       for status in ValidStatus:
#         for oldStatus in ValidStatus:
#           if status == oldStatus: continue
#           self.mock_p.evaluate.return_value = [{'Status': status, 'Reason': 'testReason', 'PolicyName': 'test_P'}]
#           pdp = PDP(self.VO, granularity, 'XX', oldStatus, None, 'XX')
#           res = pdp.takeDecision(policyIn = self.mock_p)
#           res = res['PolicyCombinedResult']
#           self.assertTrue(res['Action'])
#
#           res = pdp.takeDecision(policyIn = self.mock_p, argsIn = ())
#           res = res['PolicyCombinedResult']
#           self.assertTrue(res['Action'])
#
#           res = pdp.takeDecision(policyIn = self.mock_p, knownInfo={})
#           res = res['PolicyCombinedResult']
#           self.assertTrue(res['Action'])
#
#   def test__policyCombination(self):
#
#     for granularity in ValidRes:
#       for status in ValidStatus:
#         for oldStatus in ValidStatus:
#           if status == oldStatus:
#             continue
#
#           for newStatus1 in ValidStatus:
#             for newStatus2 in ValidStatus:
#               pdp = PDP(self.VO, granularity, 'XX', status, oldStatus, 'XX')
#               polRes  = {'Status':newStatus1, 'Reason':'-Reason1-'}
#               polRes2 = {'Status':newStatus2, 'Reason':'-Reason2-'}
#
#
#               # 0 policies
#               res = pdp._policyCombination([])
#               self.assertEqual(res, {})
#
#               # 1 policy
#               res = pdp._policyCombination([polRes])
#
#               if status == 'Banned':
#                 self.assertTrue(value_of_status(res['Status']) <= 1)
#
#               if status == 'Banned' and newStatus1 in ['Active','Bad','Probing']:
#                 self.assertEqual(res['Status'], 'Probing')
#               else:
#                 self.assertEqual(res['Status'], newStatus1)
#
#
#               # 2 policies
#               res = pdp._policyCombination([polRes, polRes2])
#
#               if status == 'Banned':
#                 self.assertTrue(value_of_status(res['Status']) <= 1)
#
#               if status == 'Banned' and newStatus1 in ['Active','Bad','Probing'] and newStatus2 in ['Active','Bad','Probing']:
#                 self.assertEqual(res['Status'], 'Probing')
#
#               if status != 'Banned' and value_of_status(newStatus1) < value_of_status(newStatus1):
#                 self.assertEqual(res['Status'], newStatus1)
#               if status != 'Banned' and value_of_status(newStatus2) < value_of_status(newStatus1):
#                 self.assertEqual(res['Status'], newStatus2)
#
#               # all different policies
#               def make_polres(status):
#                 return { 'Status': status, 'Reason': 'Because of ' + status }
#               all_polres = [make_polres(s) for s in ValidStatus]
#
#               res = pdp._policyCombination(all_polres)
#               self.assertEqual(res['Status'], 'Banned')
#
# #############################################################################
#
# class PDPFailure(PolicySystemTestCase):
#
# #############################################################################
#
#   def test_PolicyFail(self):
#     for granularity in ValidRes:
#       for status in ValidStatus:
#         for oldStatus in ValidStatus:
#           if status == oldStatus:
#             continue
#           pdp = PDP(self.VO, granularity, 'XX', status, oldStatus, 'XX')
#           self.assertRaises(Exception, pdp.takeDecision, self.mock_pdp, self.mock_rsDB)
#           self.assertRaises(Exception, pdp.takeDecision, self.mock_pdp, self.mock_rsDB, knownInfo={'DT':'AT_RISK'})
#
# #############################################################################
#
#   def test_PDPBadInputs(self):
#     for status in ValidStatus:
#       for oldStatus in ValidStatus:
#         if status == oldStatus:
#           continue
#     for granularity in ValidRes:
#       for oldStatus in ValidStatus:
#         for status in ValidStatus:
#           if status == oldStatus:
#             continue
#
# #############################################################################
#
# class PolicyCallerSuccess(PolicySystemTestCase):
#
#   def test_policyInvocation(self):
#     cc = Mock()
#
#     policies_modules = {'Site':['DT_Policy', 'GGUSTickets_Policy'],
#                         'Service': ['PilotsEfficiency_Simple_Policy', 'JobsEfficiency_Simple_Policy'],
#                         'Resource':['SAMResults_Policy', 'DT_Policy'],
#                         'StorageElementRead':['SEOccupancy_Policy', 'TransferQuality_Policy'],
#                         'StorageElementWrite':['SEOccupancy_Policy', 'TransferQuality_Policy']
#                         }
#
#     for g in ValidRes:
#       for status in ValidStatus:
#         self.mock_p.evaluate.return_value = {'Status':status,
#                                              'Reason':'testReason',
#                                              'PolicyName': 'test_P'}
#         pc = PolicyCaller(commandCallerIn = cc)
#
#         for pol_mod in policies_modules[g]:
#           res = pc.policyInvocation(self.VO, g, 'XX', status, self.mock_p,
#                                     (g, 'XX'), None, pol_mod)
#           self.assertEqual(res['Status'], status)
#
#           res = pc.policyInvocation(self.VO, g, 'XX', status, self.mock_p,
#                                     None, None, pol_mod)
#           self.assertEqual(res['Status'], status)
#
#           for extraArgs in ((g, 'XX'), [(g, 'XX'), (g, 'XX')]):
#             res = pc.policyInvocation(self.VO, g, 'XX', status, self.mock_p,
#                                       None, None, pol_mod, extraArgs)
#             self.assertEqual(res['Status'], status)
#
# #############################################################################
#
# class PolicyBaseSuccess(PolicySystemTestCase):
#
#   def test_setArgs(self):
#     for g in ValidRes:
#       for a in [(g, 'XX')]:
#         self.pb.setArgs(a)
#         self.assertEqual(self.pb.args, a)
#
#   def test_evaluate(self):
#     for g in ValidRes:
#       for a in [(g, 'XX')]:
#         self.pb.setArgs(a)
#         self.mock_command.doCommand.return_value = {'Result':'aRes'}
#         self.pb.setCommand(self.mock_command)
#         res = self.pb.evaluate()
#         self.assertEqual(res, 'aRes')
#
# #############################################################################
#
# class PolicyBaseFailure(PolicySystemTestCase):
#
#   def test_setBadArgs(self):
#
#
#     # 6 arguments should be handled with no problem: why the limitation to 5 ?! (removing this test)
#     # self.pb.setArgs(('Site', 'XX', 'Active', 'BOH', 'BOH', 'BOH'))
#     # self.mock_command.doCommand.return_value = {'Result':'aRes'}
#     # self.pb.setCommand(self.mock_command)
#     # Lists are unsupported by Command for now, useless to test.
#     # self.pb.setArgs([('Site', 'XX', 'Active', 'BOH', 'BOH', 'BOH'), ('Site', 'XX', 'Active', 'BOH', 'BOH', 'BOH')])
#
# #############################################################################
#
# # class PolicyInvokerSuccess(PolicySystemTestCase):
#
# #   def test_setPolicy(self):
# #     self.pi.setPolicy(self.mock_policy)
# #     self.assertEqual(self.pi.policy, self.mock_policy)
#
# #   def test_evaluatePolicy(self):
#
# #     self.mock_policy.evaluate.return_value = {'Result':'Satisfied', 'Status':'Banned', 'Reason':"reason"}
# #     self.pi.setPolicy(self.mock_policy)
# #     for granularity in ValidRes:
# #       res = self.pi.evaluatePolicy()
# #       self.assertEqual(res['Result'], 'Satisfied')
# #     self.mock_policy.evaluate.return_value = {'Result':'Un-Satisfied'}
# #     self.pi.setPolicy(self.mock_policy)
# #     for granularity in ValidRes:
# #       res = self.pi.evaluatePolicy()
# #       self.assertEqual(res['Result'], 'Un-Satisfied')
#
# # #############################################################################
#
# # class PolicyInvokerFailure(PolicySystemTestCase):
#
# #   def test_policyFail(self):
# #     for granularity in ValidRes:
# #       self.assertRaises(Exception, self.pi.evaluatePolicy)

#############################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(PolicySystemTestCase)
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PolicyBaseSuccess))
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PolicyBaseFailure))
    #  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PolicyInvokerSuccess))
    #  suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PolicyInvokerFailure))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PEPSuccess))
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PEPFailure))
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PDPSuccess))
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PDPFailure))
    #   suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(PolicyCallerSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

#############################################################################



""" NotificationDB class is a front-end to the Notifications database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.Mail import Mail
from DIRAC.Core.Base.DB import DB
from DIRAC.Core.Utilities import DEncode
from DIRAC.ConfigurationSystem.Client.Helpers import Registry


class NotificationDB(DB):
    def __init__(self):
        DB.__init__(self, "NotificationDB", "Framework/NotificationDB")
        result = self.__initializeDB()
        if not result["OK"]:
            self.log.fatal("Cannot initialize DB!", result["Message"])
        self.__alarmQueryFields = [
            "alarmid",
            "author",
            "creationtime",
            "modtime",
            "subject",
            "status",
            "priority",
            "notifications",
            "body",
            "assignee",
            "alarmkey",
        ]
        self.__alarmLogFields = ["timestamp", "author", "comment", "modifications"]
        self.__notificationQueryFields = ("id", "user", "seen", "message", "timestamp")
        self.__newAlarmMandatoryFields = [
            "author",
            "subject",
            "status",
            "notifications",
            "body",
            "assignee",
            "priority",
        ]
        self.__updateAlarmIdentificationFields = ["id", "alarmKey"]
        self.__updateAlarmMandatoryFields = ["author"]
        self.__updateAlarmAtLeastOneField = ["comment", "modifications"]
        self.__updateAlarmModificableFields = ["status", "assignee", "priority"]
        self.__validAlarmStatus = ["Open", "OnGoing", "Closed", "Testing"]
        self.__validAlarmNotifications = ["Web", "Mail", "SMS"]
        self.__validAlarmPriorities = ["Low", "Medium", "High", "Extreme"]

    def __initializeDB(self):
        retVal = self._query("show tables")
        if not retVal["OK"]:
            return retVal

        tablesInDB = [t[0] for t in retVal["Value"]]
        tablesToCreate = {}
        if "ntf_Alarms" not in tablesInDB:
            tablesToCreate["ntf_Alarms"] = {
                "Fields": {
                    "AlarmId": "INTEGER UNSIGNED AUTO_INCREMENT NOT NULL",
                    "AlarmKey": "VARCHAR(32) NOT NULL",
                    "Author": "VARCHAR(64) NOT NULL",
                    "CreationTime": "DATETIME NOT NULL",
                    "ModTime": "DATETIME NOT NULL",
                    "Subject": "VARCHAR(255) NOT NULL",
                    "Status": "VARCHAR(64) NOT NULL",
                    "Priority": "VARCHAR(32) NOT NULL",
                    "Body": "BLOB",
                    "Assignee": "VARCHAR(64) NOT NULL",
                    "Notifications": "VARCHAR(128) NOT NULL",
                },
                "PrimaryKey": "AlarmId",
                "Indexes": {"Status": ["Status"], "Assignee": ["Assignee"]},
            }
        if "ntf_AssigneeGroups" not in tablesInDB:
            tablesToCreate["ntf_AssigneeGroups"] = {
                "Fields": {
                    "AssigneeGroup": "VARCHAR(64) NOT NULL",
                    "User": "VARCHAR(64) NOT NULL",
                },
                "Indexes": {"ag": ["AssigneeGroup"]},
            }

        if "ntf_AlarmLog" not in tablesInDB:
            tablesToCreate["ntf_AlarmLog"] = {
                "Fields": {
                    "AlarmId": "INTEGER UNSIGNED NOT NULL",
                    "Timestamp": "DATETIME NOT NULL",
                    "Author": "VARCHAR(64) NOT NULL",
                    "Comment": "BLOB",
                    "Modifications": "VARCHAR(255)",
                },
                "Indexes": {"AlarmID": ["AlarmId"]},
            }

        if "ntf_AlarmFollowers" not in tablesInDB:
            tablesToCreate["ntf_AlarmFollowers"] = {
                "Fields": {
                    "AlarmId": "INTEGER UNSIGNED NOT NULL",
                    "User": "VARCHAR(64) NOT NULL",
                    "Mail": "TINYINT(1) DEFAULT 0",
                    "Notification": "TINYINT(1) DEFAULT 1",
                    "SMS": "TINYINT(1) DEFAULT 0",
                },
                "Indexes": {"AlarmID": ["AlarmId"]},
            }

        if "ntf_Notifications" not in tablesInDB:
            tablesToCreate["ntf_Notifications"] = {
                "Fields": {
                    "Id": "INTEGER UNSIGNED AUTO_INCREMENT NOT NULL",
                    "User": "VARCHAR(64) NOT NULL",
                    "Message": "BLOB NOT NULL",
                    "Seen": "TINYINT(1) NOT NULL DEFAULT 0",
                    "Expiration": "DATETIME",
                    "Timestamp": "DATETIME",
                    "DeferToMail": "TINYINT(1) NOT NULL DEFAULT 1",
                },
                "PrimaryKey": "Id",
            }

        if tablesToCreate:
            return self._createTables(tablesToCreate)
        return S_OK()

    def __checkAlarmField(self, name, value):
        name = name.lower()
        if name == "status":
            if value not in self.__validAlarmStatus:
                return S_ERROR("Status %s is invalid. Valid ones are: %s" % (value, self.__validAlarmStatus))
        elif name == "priority":
            if value not in self.__validAlarmPriorities:
                return S_ERROR("Type %s is invalid. Valid ones are: %s" % (value, self.__validAlarmPriorities))
        elif name == "assignee":
            result = self.getUserAsignees(value)
            if not result["OK"]:
                return result
            if not result["Value"]:
                return S_ERROR("%s is not a known assignee" % value)
            return result
        return S_OK()

    def newAlarm(self, alarmDef):
        """Create a new alarm record"""
        followers = []
        for field in self.__newAlarmMandatoryFields:
            if field not in alarmDef:
                return S_ERROR("Oops. Missing %s" % field)
            result = self.__checkAlarmField(field, alarmDef[field])
            if not result["OK"]:
                return result
            if field == "assignee":
                followers = result["Value"]
        author = alarmDef["author"]
        if author not in followers:
            followers.append(author)

        sqlFieldsName = []
        sqlFieldsValue = []
        for field in self.__newAlarmMandatoryFields:
            if field == "notifications":
                notifications = {}
                for nType in self.__validAlarmNotifications:
                    if nType in alarmDef[field]:
                        notifications[nType] = 1
                    else:
                        notifications[nType] = 0
                val = DEncode.encode(notifications)
            else:
                val = alarmDef[field]
            # Add to the list of fields to add
            sqlFieldsName.append(field)
            result = self._escapeString(val)
            if result["OK"]:
                sqlFieldsValue.append(result["Value"])
            else:
                return S_ERROR("Failed to escape value %s" % val)
        sqlFieldsName.extend(["CreationTime", "ModTime"])
        sqlFieldsValue.extend(["UTC_TIMESTAMP()", "UTC_TIMESTAMP()"])

        # Get the defined alarmkey and generate a random one if not defined
        if "alarmKey" in alarmDef:
            result = self._escapeString(alarmDef["alarmKey"])
            if result["OK"]:
                alarmKey = result["Value"]
            else:
                return S_ERROR("Failed to escape value %s for key AlarmKey" % val)
            gLogger.info("Checking there are no alarms with key %s" % alarmKey)
            result = self._query("SELECT AlarmId FROM `ntf_Alarms` WHERE AlarmKey=%s" % alarmKey)
            if not result["OK"]:
                return result
            if result["Value"]:
                return S_ERROR("Oops, alarm with id %s has the same alarm key!" % result["Value"][0][0])
        else:
            alarmKey = str(time.time())[-31:]
        sqlFieldsName.append("AlarmKey")
        sqlFieldsValue.append(alarmKey)

        sqlInsert = "INSERT INTO `ntf_Alarms` (%s) VALUES (%s)" % (",".join(sqlFieldsName), ",".join(sqlFieldsValue))

        result = self._update(sqlInsert)
        if not result["OK"]:
            return result
        alarmId = result["lastRowId"]
        for follower in followers:
            result = self.modifyFollowerForAlarm(alarmId, follower, notifications)
            if not result["OK"]:
                varMsg = "\nFollower: %s\nAlarm: %s\nError: %s" % (follower, alarmId, result["Message"])
                self.log.error("Couldn't set follower for alarm", varMsg)
        self.__notifyAlarm(alarmId)
        return S_OK(alarmId)

    def deleteAlarmsByAlarmKey(self, alarmKeyList):
        alarmsIdList = []
        for alarmKey in alarmKeyList:
            result = self.__getAlarmIdFromKey(alarmKey)
            if not result["OK"]:
                return result
            alarmId = result["Value"]
            alarmsIdList.append(alarmId)
        self.log.info("Trying to delete alarms with:\n alamKey %s\n  alarmId %s" % (alarmKeyList, alarmsIdList))
        return self.deleteAlarmsByAlarmId(alarmsIdList)

    def deleteAlarmsByAlarmId(self, alarmIdList):
        self.log.info("Trying to delete alarms with ids %s" % alarmIdList)
        try:
            alarmId = int(alarmIdList)
            alarmIdList = [alarmId]
        except Exception:
            pass

        try:
            alarmIdList = [int(alarmId) for alarmId in alarmIdList]
        except Exception:
            self.log.error("At least one alarmId is not a number", str(alarmIdList))
            return S_ERROR("At least one alarmId is not a number: %s" % str(alarmIdList))

        tablesToCheck = ("ntf_AlarmLog", "ntf_AlarmFollowers", "ntf_Alarms")
        alamsSQLList = ",".join(["%d" % alarmId for alarmId in alarmIdList])
        for tableName in tablesToCheck:
            delSql = "DELETE FROM `%s` WHERE AlarmId in ( %s )" % (tableName, alamsSQLList)
            result = self._update(delSql)
            if not result["OK"]:
                self.log.error("Could not delete alarm", "from table %s: %s" % (tableName, result["Message"]))
        return S_OK()

    def __processUpdateAlarmModifications(self, modifications):
        if not isinstance(modifications, dict):
            return S_ERROR("Modifications must be a dictionary")
        updateFields = []
        followers = []
        for field in modifications:
            if field not in self.__updateAlarmModificableFields:
                return S_ERROR("%s is not a valid modificable field" % field)
            value = modifications[field]
            result = self.__checkAlarmField(field, value)
            if not result["OK"]:
                return result
            if field == "assignee":
                followers = result["Value"]
            result = self._escapeString(modifications[field])
            if not result["OK"]:
                return result
            updateFields.append("%s=%s" % (field, result["Value"]))
        return S_OK((", ".join(updateFields), DEncode.encode(modifications), followers))

    def __getAlarmIdFromKey(self, alarmKey):
        result = self._escapeString(alarmKey)
        if not result["OK"]:
            return S_ERROR("Cannot escape alarmKey %s" % alarmKey)
        alarmKey = result["Value"]
        sqlQuery = "SELECT AlarmId FROM `ntf_Alarms` WHERE AlarmKey=%s" % alarmKey
        result = self._query(sqlQuery)
        if result["OK"]:
            result["Value"] = result["Value"][0][0]
        return result

    def updateAlarm(self, updateReq):
        # Discover alarm identification
        idOK = False
        for field in self.__updateAlarmIdentificationFields:
            if field in updateReq:
                idOK = True
        if not idOK:
            return S_ERROR(
                "Need at least one field to identify which alarm to update! %s" % self.__updateAlarmIdentificationFields
            )
        if "alarmKey" in updateReq:
            alarmKey = updateReq["alarmKey"]
            result = self.__getAlarmIdFromKey(alarmKey)
            if not result["OK"]:
                self.log.error("Could not get alarm id for key", " %s: %s" % (alarmKey, result["Value"]))
                return result
            updateReq["id"] = result["Value"]
            self.log.info("Retrieving alarm key %s maps to id %s" % (alarmKey, updateReq["id"]))
        # Check fields
        for field in self.__updateAlarmMandatoryFields:
            if field not in updateReq:
                return S_ERROR("Oops. Missing %s" % field)
        validReq = False
        for field in self.__updateAlarmAtLeastOneField:
            if field in updateReq:
                validReq = True
        if not validReq:
            return S_OK("Requirement needs at least one of %s" % " ".join(self.__updateAlarmAtLeastOneField))
        author = updateReq["author"]
        followers = [author]
        if author not in Registry.getAllUsers():
            return S_ERROR("%s is not a known user" % author)
        result = self._escapeString(author)
        if not result["OK"]:
            return result
        author = result["Value"]
        try:
            alarmId = int(updateReq["id"])
        except Exception:
            return S_ERROR("Oops, Alarm id is not valid!")
        result = self._query("SELECT AlarmId FROM `ntf_Alarms` WHERE AlarmId=%d" % alarmId)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Alarm %s does not exist!" % alarmId)
        sqlFields = ["AlarmId", "Author", "Timestamp"]
        sqlValues = ["%d" % alarmId, author, "UTC_TIMESTAMP()"]
        rawComment = ""
        if "comment" in updateReq:
            rawComment = updateReq["comment"]
            result = self._escapeString(rawComment)
            if not result["OK"]:
                return result
            sqlFields.append("Comment")
            sqlValues.append(result["Value"])
        modifications = False
        if "modifications" in updateReq:
            modifications = updateReq["modifications"]
            result = self.__processUpdateAlarmModifications(modifications)
            if not result["OK"]:
                return result
            alarmModsSQL, encodedMods, newFollowers = result["Value"]
            sqlFields.append("Modifications")
            result = self._escapeString(encodedMods)
            if not result["OK"]:
                return result
            sqlValues.append(result["Value"])
            if newFollowers:
                followers.extend(newFollowers)
        logSQL = "INSERT INTO `ntf_AlarmLog` (%s) VALUES (%s)" % (",".join(sqlFields), ",".join(sqlValues))
        result = self._update(logSQL)
        if not result["OK"]:
            return result
        modSQL = "ModTime=UTC_TIMESTAMP()"
        if modifications:
            modSQL = "%s, %s" % (modSQL, alarmModsSQL)
        updateSQL = "UPDATE `ntf_Alarms` SET %s WHERE AlarmId=%d" % (modSQL, alarmId)
        result = self._update(updateSQL)
        if not result["OK"]:
            return result
        # Get notifications config
        sqlQuery = "SELECT Notifications FROM `ntf_Alarms` WHERE AlarmId=%s" % alarmId
        result = self._query(sqlQuery)
        if not result["OK"] or not result["Value"]:
            self.log.error("Could not retrieve default notifications for alarm", "%s" % alarmId)
            return S_OK(alarmId)
        notificationsDict = DEncode.decode(result["Value"][0][0])[0]
        for v in self.__validAlarmNotifications:
            if v not in notificationsDict:
                notificationsDict[v] = 0
        for follower in followers:
            result = self.modifyFollowerForAlarm(alarmId, follower, notificationsDict, overwrite=False)
            if not result["OK"]:
                varMsg = "\nFollower: %s\nAlarm: %s\nError: %s" % (follower, alarmId, result["Message"])
                self.log.error("Couldn't set follower for alarm", varMsg)
        return self.__notifyAlarm(alarmId)

    def __notifyAlarm(self, alarmId):
        result = self.getSubscribersForAlarm(alarmId)
        if not result["OK"]:
            return result
        subscribers = result["Value"]
        needLongText = False
        if subscribers["mail"]:
            needLongText = True
        result = self.getAlarmInfo(alarmId)
        if not result["OK"]:
            return result
        alarmInfo = result["Value"]
        result = self.getAlarmLog(alarmId)
        if not result["OK"]:
            return result
        alarmLog = result["Value"]
        if subscribers["notification"]:
            msg = self.__generateAlarmInfoMessage(alarmInfo)
            logMsg = self.__generateAlarmLogMessage(alarmLog, True)
            if logMsg:
                msg = "%s\n\n%s\nLast modification:\n%s" % (msg, "*" * 30, logMsg)
            for user in subscribers["notification"]:
                self.addNotificationForUser(user, msg, 86400, deferToMail=True)
        if subscribers["mail"]:
            msg = self.__generateAlarmInfoMessage(alarmInfo)
            logMsg = self.__generateAlarmLogMessage(alarmLog)
            if logMsg:
                msg = "%s\n\n%s\nAlarm Log:\n%s" % (msg, "*" * 30, logMsg)
                subject = "Update on alarm %s" % alarmId
            else:
                subject = "New alarm %s" % alarmId
            for user in subscribers["mail"]:
                self.__sendMailToUser(user, subject, msg)
        if subscribers["sms"]:
            # TODO
            pass
        return S_OK()

    def __generateAlarmLogMessage(self, alarmLog, showOnlyLast=False):
        if len(alarmLog["Records"]) == 0:
            return ""
        records = alarmLog["Records"]
        if showOnlyLast:
            logToShow = [-1]
        else:
            logToShow = list(range(len(records) - 1, -1, -1))
        finalMessage = []
        for iD in logToShow:
            rec = records[iD]
            data = {}
            for i in range(len(alarmLog["ParameterNames"])):
                if rec[i]:
                    data[alarmLog["ParameterNames"][i]] = rec[i]
            # [ 'timestamp', 'author', 'comment', 'modifications' ]
            msg = [" Entry by : %s" % data["author"]]
            msg.append(" On       : %s" % data["timestamp"].strftime("%Y/%m/%d %H:%M:%S"))
            if "modifications" in data:
                mods = data["modifications"]
                keys = sorted(mods)
                msg.append(" Modificaitons:")
                for key in keys:
                    msg.append("   %s -> %s" % (key, mods[key]))
            if "comment" in data:
                msg.append(" Comment:\n\n%s" % data["comment"])
            finalMessage.append("\n".join(msg))
        return "\n\n===============\n".join(finalMessage)

    def __generateAlarmInfoMessage(self, alarmInfo):
        # [ 'alarmid', 'author', 'creationtime', 'modtime', 'subject', 'status', 'type', 'body', 'assignee' ]
        msg = " Alarm %6d\n" % alarmInfo["alarmid"]
        msg += "   Author            : %s\n" % alarmInfo["author"]
        msg += "   Subject           : %s\n" % alarmInfo["subject"]
        msg += "   Status            : %s\n" % alarmInfo["status"]
        msg += "   Priority          : %s\n" % alarmInfo["priority"]
        msg += "   Assignee          : %s\n" % alarmInfo["assignee"]
        msg += "   Creation date     : %s UTC\n" % alarmInfo["creationtime"].strftime("%Y/%m/%d %H:%M:%S")
        msg += "   Last modificaiton : %s UTC\n" % alarmInfo["modtime"].strftime("%Y/%m/%d %H:%M:%S")
        msg += "   Body:\n\n%s" % alarmInfo["body"]
        return msg

    def __sendMailToUser(self, user, subject, message):
        address = gConfig.getValue("/Registry/Users/%s/Email" % user, "")
        if not address:
            self.log.error("User does not have an email registered", user)
            return S_ERROR("User %s does not have an email registered" % user)
        self.log.info("Sending mail (%s) to user %s at %s" % (subject, user, address))
        m = Mail()
        m._subject = "[DIRAC] %s" % subject
        m._message = message
        m._mailAddress = address
        result = m._send()
        if not result["OK"]:
            gLogger.warn("Could not send mail with the following message:\n%s" % result["Message"])

        return result

    def getAlarms(self, condDict={}, sortList=False, start=0, limit=0, modifiedAfter=None):

        condSQL = []
        for field in self.__alarmQueryFields:
            if field in condDict:
                fieldValues = []
                rawValue = condDict[field]
                if field == "assignee":
                    expandedValue = []
                    for user in rawValue:
                        result = self.getAssigneeGroupsForUser(user)
                        if not result["OK"]:
                            return result
                        for ag in result["Value"]:
                            if ag not in expandedValue:
                                expandedValue.append(ag)
                    rawValue = expandedValue
                for value in rawValue:
                    result = self._escapeString(value)
                    if not result["OK"]:
                        return result
                    fieldValues.append(result["Value"])
                condSQL.append("%s in ( %s )" % (field, ",".join(fieldValues)))

        selSQL = "SELECT %s FROM `ntf_Alarms`" % ",".join(self.__alarmQueryFields)
        if modifiedAfter:
            condSQL.append("ModTime >= %s" % modifiedAfter.strftime("%Y-%m-%d %H:%M:%S"))
        if condSQL:
            selSQL = "%s WHERE %s" % (selSQL, " AND ".join(condSQL))
        if sortList:
            selSQL += " ORDER BY %s" % ", ".join(["%s %s" % (sort[0], sort[1]) for sort in sortList])
        if limit:
            selSQL += " LIMIT %d,%d" % (start, limit)

        result = self._query(selSQL)
        if not result["OK"]:
            return result

        resultDict = {}
        resultDict["ParameterNames"] = self.__alarmQueryFields
        resultDict["Records"] = [list(v) for v in result["Value"]]
        return S_OK(resultDict)

    def getAlarmInfo(self, alarmId):
        result = self.getAlarms({"alarmId": alarmId})
        if not result["OK"]:
            return result
        alarmInfo = {}
        data = result["Value"]
        if len(data["Records"]) == 0:
            return S_OK({})
        for i in range(len(data["ParameterNames"])):
            alarmInfo[data["ParameterNames"][i]] = data["Records"][0][i]
        return S_OK(alarmInfo)

    def getAlarmLog(self, alarmId):
        try:
            alarmId = int(alarmId)
        except Exception:
            return S_ERROR("Alarm id must be a non decimal number")
        sqlSel = "SELECT %s FROM `ntf_AlarmLog` WHERE AlarmId=%d ORDER BY Timestamp ASC" % (
            ",".join(self.__alarmLogFields),
            alarmId,
        )
        result = self._query(sqlSel)
        if not result["OK"]:
            return result
        decodedRows = []
        for row in result["Value"]:
            decodedRows.append(list(row))
            if not row[3]:
                decodedRows.append(list(row))
                continue
            dec = DEncode.decode(row[3])
            decodedRows[-1][3] = dec[0]

        resultDict = {}
        resultDict["ParameterNames"] = self.__alarmLogFields
        resultDict["Records"] = decodedRows
        return S_OK(resultDict)

    ###
    # Followers management
    ###

    def modifyFollowerForAlarm(self, alarmId, user, notificationsDict, overwrite=True):
        rawUser = user
        if rawUser not in Registry.getAllUsers():
            return S_OK()
        result = self._escapeString(user)
        if not result["OK"]:
            return result
        user = result["Value"]
        subscriber = False
        for k in notificationsDict:
            if notificationsDict[k]:
                subscriber = True
                break
        selSQL = "SELECT Notification, Mail, SMS FROM `ntf_AlarmFollowers` WHERE AlarmId=%d AND User=%s" % (
            alarmId,
            user,
        )
        result = self._query(selSQL)
        if not result["OK"]:
            return result
        if not result["Value"]:
            if not subscriber:
                return S_OK()
            sqlValues = ["%d" % alarmId, user]
            for k in self.__validAlarmNotifications:
                if notificationsDict[k]:
                    sqlValues.append("1")
                else:
                    sqlValues.append("0")
            inSQL = (
                "INSERT INTO `ntf_AlarmFollowers` ( AlarmId, User, Notification, Mail, SMS ) VALUES (%s)"
                % ",".join(sqlValues)
            )
            return self._update(inSQL)
        sqlCond = "AlarmId=%d AND User=%s" % (alarmId, user)
        # Need to delete
        if not subscriber:
            return self._update("DELETE FROM `ntf_AlarmFollowers` WHERE %s" % sqlCond)
        if not overwrite:
            return S_OK()
        # Need to update
        modSQL = []
        for k in self.__validAlarmNotifications:
            if notificationsDict[k]:
                modSQL.append("%s=1" % k)
            else:
                modSQL.append("%s=0" % k)
        return self._update("UPDATE `ntf_AlarmFollowers` SET %s WHERE %s" % (modSQL, sqlCond))

    def getSubscribersForAlarm(self, alarmId):
        selSQL = "SELECT User, Mail, Notification, SMS FROM `ntf_AlarmFollowers` WHERE AlarmId=%d" % alarmId
        result = self._query(selSQL)
        if not result["OK"]:
            return result
        fw = result["Value"]
        followWays = {"mail": [], "notification": [], "sms": []}
        followers = []
        for user, mail, Notification, SMS in fw:
            if user in followers:
                continue
            followers.append(user)
            if mail:
                followWays["mail"].append(user)
            if Notification:
                followWays["notification"].append(user)
            if SMS:
                followWays["sms"].append(user)
        return S_OK(followWays)

    ###
    # Assignee groups management
    ###

    def getUserAsignees(self, assignee):
        # Check if it is a user
        if assignee in Registry.getAllUsers():
            return S_OK([assignee])
        result = self._escapeString(assignee)
        if not result["OK"]:
            return result
        escAG = result["Value"]
        sqlSel = "SELECT User FROM `ntf_AssigneeGroups` WHERE AssigneeGroup = %s" % escAG
        result = self._query(sqlSel)
        if not result["OK"]:
            return result
        users = [row[0] for row in result["Value"]]
        if not users:
            return S_OK([])
        return S_OK(users)

    def setAssigneeGroup(self, groupName, usersList):
        validUsers = Registry.getAllUsers()
        result = self._escapeString(groupName)
        if not result["OK"]:
            return result
        escGroup = result["Value"]
        sqlSel = "SELECT User FROM `ntf_AssigneeGroups` WHERE AssigneeGroup = %s" % escGroup
        result = self._query(sqlSel)
        if not result["OK"]:
            return result
        currentUsers = [row[0] for row in result["Value"]]
        usersToDelete = []
        usersToAdd = []
        finalUsersInGroup = len(currentUsers)
        for user in currentUsers:
            if user not in usersList:
                result = self._escapeString(user)
                if not result["OK"]:
                    return result
                usersToDelete.append(result["Value"])
                finalUsersInGroup -= 1
        for user in usersList:
            if user not in validUsers:
                continue
            if user not in currentUsers:
                result = self._escapeString(user)
                if not result["OK"]:
                    return result
                usersToAdd.append("( %s, %s )" % (escGroup, result["Value"]))
                finalUsersInGroup += 1
        if not finalUsersInGroup:
            return S_ERROR("Group must have at least one user!")
        # Delete old users
        if usersToDelete:
            sqlDel = "DELETE FROM `ntf_AssigneeGroups` WHERE User in ( %s )" % ",".join(usersToDelete)
            result = self._update(sqlDel)
            if not result["OK"]:
                return result
        # Add new users
        if usersToAdd:
            sqlInsert = "INSERT INTO `ntf_AssigneeGroups` ( AssigneeGroup, User ) VALUES %s" % ",".join(usersToAdd)
            result = self._update(sqlInsert)
            if not result["OK"]:
                return result
        return S_OK()

    def deleteAssigneeGroup(self, groupName):
        result = self._escapeString(groupName)
        if not result["OK"]:
            return result
        escGroup = result["Value"]
        sqlSel = "SELECT AlarmId FROM `ntf_Alarms` WHERE Assignee=%s" % escGroup
        result = self._query(sqlSel)
        if not result["OK"]:
            return result
        if result["Value"]:
            alarmIds = [row[0] for row in result["Value"]]
            return S_ERROR("There are %s alarms assigned to this group" % len(alarmIds))
        sqlDel = "DELETE FROM `ntf_AssigneeGroups` WHERE AssigneeGroup=%s" % escGroup
        return self._update(sqlDel)

    def getAssigneeGroups(self):
        result = self._query("SELECT AssigneeGroup, User from `ntf_AssigneeGroups` ORDER BY User")
        if not result["OK"]:
            return result
        agDict = {}
        for row in result["Value"]:
            ag = row[0]
            user = row[1]
            if ag not in agDict:
                agDict[ag] = []
            agDict[ag].append(user)
        return S_OK(agDict)

    def getAssigneeGroupsForUser(self, user):
        if user not in Registry.getAllUsers():
            return S_ERROR("%s is an unknown user" % user)
        result = self._escapeString(user)
        if not result["OK"]:
            return result
        user = result["Value"]
        result = self._query("SELECT AssigneeGroup from `ntf_AssigneeGroups` WHERE User=%s" % user)
        if not result["OK"]:
            return result
        return S_OK([row[0] for row in result["Value"]])

    ###
    # Notifications
    ###

    def addNotificationForUser(self, user, message, lifetime=0, deferToMail=1):
        if user not in Registry.getAllUsers():
            return S_ERROR("%s is an unknown user" % user)
        self.log.info("Adding a notification for user %s (msg is %s chars)" % (user, len(message)))
        result = self._escapeString(user)
        if not result["OK"]:
            return result
        user = result["Value"]
        result = self._escapeString(message)
        if not result["OK"]:
            return result
        message = result["Value"]
        sqlFields = ["User", "Message", "Timestamp"]
        sqlValues = [user, message, "UTC_TIMESTAMP()"]
        if not deferToMail:
            sqlFields.append("DeferToMail")
            sqlValues.append("0")
        if lifetime:
            sqlFields.append("Expiration")
            sqlValues.append("TIMESTAMPADD( SECOND, %d, UTC_TIMESTAMP() )" % int(lifetime))
        sqlInsert = "INSERT INTO `ntf_Notifications` (%s) VALUES (%s) " % (",".join(sqlFields), ",".join(sqlValues))
        result = self._update(sqlInsert)
        if not result["OK"]:
            return result
        return S_OK(result["lastRowId"])

    def removeNotificationsForUser(self, user, msgIds=False):
        if user not in Registry.getAllUsers():
            return S_ERROR("%s is an unknown user" % user)
        result = self._escapeString(user)
        if not result["OK"]:
            return result
        user = result["Value"]
        delSQL = "DELETE FROM `ntf_Notifications` WHERE User=%s" % user
        escapedIDs = []
        if msgIds:
            for iD in msgIds:
                result = self._escapeString(str(iD))
                if not result["OK"]:
                    return result
                escapedIDs.append(result["Value"])
            delSQL = "%s AND Id in ( %s ) " % (delSQL, ",".join(escapedIDs))
        return self._update(delSQL)

    def markNotificationsSeen(self, user, seen=True, msgIds=False):
        if user not in Registry.getAllUsers():
            return S_ERROR("%s is an unknown user" % user)
        result = self._escapeString(user)
        if not result["OK"]:
            return result
        user = result["Value"]
        if seen:
            seen = 1
        else:
            seen = 0
        updateSQL = "UPDATE `ntf_Notifications` SET Seen=%d WHERE User=%s" % (seen, user)
        escapedIDs = []
        if msgIds:
            for iD in msgIds:
                result = self._escapeString(str(iD))
                if not result["OK"]:
                    return result
                escapedIDs.append(result["Value"])
            updateSQL = "%s AND Id in ( %s ) " % (updateSQL, ",".join(escapedIDs))
        return self._update(updateSQL)

    def getNotifications(self, condDict={}, sortList=False, start=0, limit=0):

        condSQL = []
        for field in self.__notificationQueryFields:
            if field in condDict:
                fieldValues = []
                for value in condDict[field]:
                    result = self._escapeString(value)
                    if not result["OK"]:
                        return result
                    fieldValues.append(result["Value"])
                condSQL.append("%s in ( %s )" % (field, ",".join(fieldValues)))

        eSortList = []
        for field, order in sortList:
            if order.lower() in ["asc", "desc"]:
                eSortList.append(("`%s`" % field.replace("`", ""), order))

        selSQL = "SELECT %s FROM `ntf_Notifications`" % ",".join(self.__notificationQueryFields)
        if condSQL:
            selSQL = "%s WHERE %s" % (selSQL, " AND ".join(condSQL))
        if eSortList:
            selSQL += " ORDER BY %s" % ", ".join(["%s %s" % (sort[0], sort[1]) for sort in eSortList])
        else:
            selSQL += " ORDER BY Id DESC"
        if limit:
            selSQL += " LIMIT %d,%d" % (start, limit)

        result = self._query(selSQL)
        if not result["OK"]:
            return result

        resultDict = {}
        resultDict["ParameterNames"] = self.__notificationQueryFields
        resultDict["Records"] = [list(v) for v in result["Value"]]
        return S_OK(resultDict)

    def purgeExpiredNotifications(self):
        self.log.info("Purging expired notifications")
        delConds = ["(Seen=1 OR DeferToMail=0)", "(TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), Expiration ) < 0 )"]
        delSQL = "DELETE FROM `ntf_Notifications` WHERE %s" % " AND ".join(delConds)
        result = self._update(delSQL)
        if not result["OK"]:
            return result
        self.log.info("Purged %s notifications" % result["Value"])
        deferCond = ["Seen=0", "DeferToMail=1", "TIMESTAMPDIFF( SECOND, UTC_TIMESTAMP(), Expiration ) < 0"]
        selSQL = "SELECT Id, User, Message FROM `ntf_Notifications` WHERE %s" % " AND ".join(deferCond)
        result = self._query(selSQL)
        if not result["OK"]:
            return result
        messages = result["Value"]
        if not messages:
            return S_OK()
        ids = []
        for msg in messages:
            self.__sendMailToUser(msg[1], "Notification defered to mail", msg[2])
            ids.append(str(msg[0]))
        self.log.info("Deferred %s notifications" % len(ids))
        return self._update("DELETE FROM `ntf_Notifications` WHERE Id in (%s)" % ",".join(ids))



"""
  Service class implements the server side part of the DISET protocol
  There are 2 main parts in this class:

  - All useful functions for initialization
  - All useful functions to handle the requests
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: skip-file
# __searchInitFunctions gives RuntimeError: maximum recursion depth exceeded

import os
import time
import threading

from concurrent.futures import ThreadPoolExecutor

import six
import DIRAC
from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.Core.DISET.private.LockManager import LockManager
from DIRAC.Core.DISET.private.ServiceConfiguration import ServiceConfiguration
from DIRAC.Core.DISET.private.TransportPool import getGlobalTransportPool
from DIRAC.Core.DISET.private.MessageBroker import MessageBroker, MessageSender
from DIRAC.Core.DISET.AuthManager import AuthManager
from DIRAC.Core.DISET.RequestHandler import getServiceOption
from DIRAC.Core.Utilities import Time, MemStat, Network
from DIRAC.Core.Utilities.DErrno import ENOAUTH
from DIRAC.Core.Utilities.ReturnValues import isReturnStructure
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.FrameworkSystem.Client.MonitoringClient import MonitoringClient
from DIRAC.FrameworkSystem.Client.SecurityLogClient import SecurityLogClient

__RCSID__ = "$Id$"


class Service(object):

    SVC_VALID_ACTIONS = {"RPC": "export", "FileTransfer": "transfer", "Message": "msg", "Connection": "Message"}
    SVC_SECLOG_CLIENT = SecurityLogClient()

    def __init__(self, serviceData):
        """
        Init the variables for the service

        :param serviceData: dict with modName, standalone, loadName, moduleObj, classObj. e.g.:
          {'modName': 'Framework/serviceName',
          'standalone': True,
          'loadName': 'Framework/serviceName',
          'moduleObj': <module 'serviceNameHandler' from '/home/DIRAC/FrameworkSystem/Service/serviceNameHandler.pyo'>,
          'classObj': <class 'serviceNameHandler.serviceHandler'>}

          Standalone is true if there is only one service started
          If it's false, every service is linked to a different MonitoringClient
        """
        self._svcData = serviceData
        self._name = serviceData["modName"]
        self._startTime = Time.dateTime()
        self._validNames = [serviceData["modName"]]
        if serviceData["loadName"] not in self._validNames:
            self._validNames.append(serviceData["loadName"])
        self._cfg = ServiceConfiguration(list(self._validNames))
        self._standalone = serviceData["standalone"]
        self.__monitorLastStatsUpdate = time.time()
        self._stats = {"queries": 0, "connections": 0}
        self._authMgr = AuthManager("%s/Authorization" % PathFinder.getServiceSection(serviceData["loadName"]))
        self._transportPool = getGlobalTransportPool()
        self.__cloneId = 0
        self.__maxFD = 0

    def setCloneProcessId(self, cloneId):
        self.__cloneId = cloneId
        if not self.activityMonitoring:
            self._monitor.setComponentName("%s-Clone:%s" % (self._name, cloneId))

    def _isMetaAction(self, action):
        referedAction = Service.SVC_VALID_ACTIONS[action]
        if referedAction in Service.SVC_VALID_ACTIONS:
            return referedAction
        return False

    def initialize(self):
        # Build the URLs
        self._url = self._cfg.getURL()
        if not self._url:
            return S_ERROR("Could not build service URL for %s" % self._name)
        gLogger.verbose("Service URL is %s" % self._url)
        # Load handler
        result = self._loadHandlerInit()
        if not result["OK"]:
            return result
        self._handler = result["Value"]
        # Initialize lock manager
        self._lockManager = LockManager(self._cfg.getMaxWaitingPetitions())
        self._threadPool = ThreadPoolExecutor(max(0, self._cfg.getMaxThreads()))
        self._msgBroker = MessageBroker("%sMSB" % self._name, threadPool=self._threadPool)
        # Create static dict
        self._serviceInfoDict = {
            "serviceName": self._name,
            "serviceSectionPath": PathFinder.getServiceSection(self._name),
            "URL": self._cfg.getURL(),
            "messageSender": MessageSender(self._name, self._msgBroker),
            "validNames": self._validNames,
            "csPaths": [PathFinder.getServiceSection(svcName) for svcName in self._validNames],
        }
        # Initialize Monitoring
        # This is a flag used to check whether "EnableActivityMonitoring" is enabled or not from the config file.
        self.activityMonitoring = Operations().getValue("EnableActivityMonitoring", False) or getServiceOption(
            self._serviceInfoDict, "EnableActivityMonitoring", False
        )
        if self.activityMonitoring:
            # The import needs to be here because of the CS must be initialized before importing
            # this class (see https://github.com/DIRACGrid/DIRAC/issues/4793)
            from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

            self.activityMonitoringReporter = MonitoringReporter(monitoringType="ComponentMonitoring")
            gThreadScheduler.addPeriodicTask(100, self.__activityMonitoringReporting)
        elif self._standalone:
            self._monitor = gMonitor
        else:
            self._monitor = MonitoringClient()
        self._initMonitoring()
        # Call static initialization function
        try:
            if self.activityMonitoring:
                self._handler["class"]._rh__initializeClass(
                    dict(self._serviceInfoDict), self._lockManager, self._msgBroker, self.activityMonitoringReporter
                )
            else:
                self._handler["class"]._rh__initializeClass(
                    dict(self._serviceInfoDict), self._lockManager, self._msgBroker, self._monitor
                )
            if self._handler["init"]:
                for initFunc in self._handler["init"]:
                    gLogger.verbose("Executing initialization function")
                    try:
                        result = initFunc(dict(self._serviceInfoDict))
                    except Exception as excp:
                        gLogger.exception("Exception while calling initialization function", lException=excp)
                        return S_ERROR("Exception while calling initialization function: %s" % str(excp))
                    if not isReturnStructure(result):
                        return S_ERROR("Service initialization function %s must return S_OK/S_ERROR" % initFunc)
                    if not result["OK"]:
                        return S_ERROR("Error while initializing %s: %s" % (self._name, result["Message"]))
        except Exception as e:
            errMsg = "Exception while initializing %s" % self._name
            gLogger.exception(e)
            gLogger.exception(errMsg)
            return S_ERROR(errMsg)

        # Load actions after the handler has initialized itself
        result = self._loadActions()
        if not result["OK"]:
            return result
        self._actions = result["Value"]

        if not self.activityMonitoring:
            gThreadScheduler.addPeriodicTask(30, self.__reportThreadPoolContents)

        return S_OK()

    def __searchInitFunctions(self, handlerClass, currentClass=None):
        if not currentClass:
            currentClass = handlerClass
        initFuncs = []
        ancestorHasInit = False
        for ancestor in currentClass.__bases__:
            initFuncs += self.__searchInitFunctions(handlerClass, ancestor)
            if "initializeHandler" in dir(ancestor):
                ancestorHasInit = True
        if ancestorHasInit:
            initFuncs.append(super(currentClass, handlerClass).initializeHandler)
        if currentClass == handlerClass and "initializeHandler" in dir(handlerClass):
            initFuncs.append(handlerClass.initializeHandler)
        return initFuncs

    def _loadHandlerInit(self):
        handlerClass = self._svcData["classObj"]
        handlerName = handlerClass.__name__
        handlerInitMethods = self.__searchInitFunctions(handlerClass)
        try:
            handlerInitMethods.append(getattr(self._svcData["moduleObj"], "initialize%s" % handlerName))
        except AttributeError:
            gLogger.verbose("Not found global initialization function for service")

        if handlerInitMethods:
            gLogger.info("Found %s initialization methods" % len(handlerInitMethods))

        handlerInfo = {}
        handlerInfo["name"] = handlerName
        handlerInfo["module"] = self._svcData["moduleObj"]
        handlerInfo["class"] = handlerClass
        handlerInfo["init"] = handlerInitMethods

        return S_OK(handlerInfo)

    def _loadActions(self):

        handlerClass = self._handler["class"]

        authRules = {}
        typeCheck = {}
        methodsList = {}
        for actionType in Service.SVC_VALID_ACTIONS:
            if self._isMetaAction(actionType):
                continue
            authRules[actionType] = {}
            typeCheck[actionType] = {}
            methodsList[actionType] = []
        handlerAttributeList = dir(handlerClass)
        for actionType in Service.SVC_VALID_ACTIONS:
            if self._isMetaAction(actionType):
                continue
            methodPrefix = "%s_" % Service.SVC_VALID_ACTIONS[actionType]
            for attribute in handlerAttributeList:
                if attribute.find(methodPrefix) != 0:
                    continue
                exportedName = attribute[len(methodPrefix) :]
                methodsList[actionType].append(exportedName)
                gLogger.verbose("+ Found %s method %s" % (actionType, exportedName))
                # Create lock for method
                self._lockManager.createLock(
                    "%s/%s" % (actionType, exportedName), self._cfg.getMaxThreadsForMethod(actionType, exportedName)
                )
                # Look for type and auth rules
                if actionType == "RPC":
                    typeAttr = "types_%s" % exportedName
                    authAttr = "auth_%s" % exportedName
                else:
                    typeAttr = "types_%s_%s" % (Service.SVC_VALID_ACTIONS[actionType], exportedName)
                    authAttr = "auth_%s_%s" % (Service.SVC_VALID_ACTIONS[actionType], exportedName)
                if typeAttr in handlerAttributeList:
                    obj = getattr(handlerClass, typeAttr)
                    gLogger.verbose("|- Found type definition %s: %s" % (typeAttr, str(obj)))
                    typeCheck[actionType][exportedName] = obj
                if authAttr in handlerAttributeList:
                    obj = getattr(handlerClass, authAttr)
                    gLogger.verbose("|- Found auth rules %s: %s" % (authAttr, str(obj)))
                    authRules[actionType][exportedName] = obj

        for actionType in Service.SVC_VALID_ACTIONS:
            referedAction = self._isMetaAction(actionType)
            if not referedAction:
                continue
            gLogger.verbose("Action %s is a meta action for %s" % (actionType, referedAction))
            authRules[actionType] = []
            for method in authRules[referedAction]:
                for prop in authRules[referedAction][method]:
                    if prop not in authRules[actionType]:
                        authRules[actionType].append(prop)
            gLogger.verbose("Meta action %s props are %s" % (actionType, authRules[actionType]))

        return S_OK({"methods": methodsList, "auth": authRules, "types": typeCheck})

    def _initMonitoring(self):
        if not self.activityMonitoring:
            # Init extra bits of monitoring
            self._monitor.setComponentType(MonitoringClient.COMPONENT_SERVICE)
            self._monitor.setComponentName(self._name)
            self._monitor.setComponentLocation(self._cfg.getURL())
            self._monitor.initialize()
            self._monitor.registerActivity(
                "Connections", "Connections received", "Framework", "connections", MonitoringClient.OP_RATE
            )
            self._monitor.registerActivity(
                "Queries", "Queries served", "Framework", "queries", MonitoringClient.OP_RATE
            )
            self._monitor.registerActivity("CPU", "CPU Usage", "Framework", "CPU,%", MonitoringClient.OP_MEAN, 600)
            self._monitor.registerActivity(
                "MEM", "Memory Usage", "Framework", "Memory,MB", MonitoringClient.OP_MEAN, 600
            )
            self._monitor.registerActivity(
                "PendingQueries", "Pending queries", "Framework", "queries", MonitoringClient.OP_MEAN
            )
            self._monitor.registerActivity(
                "ActiveQueries", "Active queries", "Framework", "threads", MonitoringClient.OP_MEAN
            )
            self._monitor.registerActivity(
                "RunningThreads", "Running threads", "Framework", "threads", MonitoringClient.OP_MEAN
            )
            self._monitor.registerActivity("MaxFD", "Max File Descriptors", "Framework", "fd", MonitoringClient.OP_MEAN)

            self._monitor.setComponentExtraParam("DIRACVersion", DIRAC.version)
            self._monitor.setComponentExtraParam("platform", DIRAC.getPlatform())
            self._monitor.setComponentExtraParam("startTime", Time.dateTime())
            props = [("__doc__", "description")]
            if six.PY2:
                props += [("__RCSID__", "version")]
            for prop in props:
                try:
                    value = getattr(self._handler["module"], prop[0])
                except Exception as e:
                    gLogger.exception(e)
                    gLogger.error("Missing property", prop[0])
                    value = "unset"
                self._monitor.setComponentExtraParam(prop[1], value)

        for secondaryName in self._cfg.registerAlsoAs():
            gLogger.info("Registering %s also as %s" % (self._name, secondaryName))
            self._validNames.append(secondaryName)

        return S_OK()

    def __reportThreadPoolContents(self):
        # TODO: remove later
        pendingQueries = self._threadPool._work_queue.qsize()
        activeQuereies = len(self._threadPool._threads)

        self._monitor.addMark("PendingQueries", pendingQueries)
        self._monitor.addMark("ActiveQueries", activeQuereies)
        self._monitor.addMark("RunningThreads", threading.activeCount())
        self._monitor.addMark("MaxFD", self.__maxFD)
        self.__maxFD = 0

    def getConfig(self):
        return self._cfg

    # End of initialization functions

    def handleConnection(self, clientTransport):
        """
        This method may be called by ServiceReactor.
        The method stacks openened connection in a queue, another thread
        read this queue and handle connection.

        :param clientTransport: Object which describes opened connection (PlainTransport or SSLTransport)
        """
        if not self.activityMonitoring:
            self._stats["connections"] += 1
            self._monitor.setComponentExtraParam("queries", self._stats["connections"])
        self._threadPool.submit(self._processInThread, clientTransport)

    @property
    def wantsThrottle(self):
        """Boolean property for if the service wants requests to stop being accepted"""
        nQueued = self._threadPool._work_queue.qsize()
        return nQueued > self._cfg.getMaxWaitingPetitions()

    # Threaded process function
    def _processInThread(self, clientTransport):
        """
        This method handles a RPC, FileTransfer or Connection.
        Connection may be opened via ServiceReactor.__acceptIncomingConnection


        - Do the SSL/TLS Handshake (if dips is used) and extract credentials
        - Get the action called by the client
        - Check if the client is authorized to perform ation
          - If not, connection is closed
        - Instanciate the RequestHandler (RequestHandler contain all methods callable)

        (Following is not directly in this method but it describe what happen at
        #Execute the action)
        - Notify the client we're ready to execute the action (via _processProposal)
          and call RequestHandler._rh_executeAction()
        - Receive arguments/file/something else (depending on action) in the RequestHandler
        - Executing the action asked by the client

        :param clientTransport: Object which describe the opened connection (SSLTransport or PlainTransport)

        :return: S_OK with "closeTransport" a boolean to indicate if th connection have to be closed
                e.g. after RPC, closeTransport=True

        """
        self.__maxFD = max(self.__maxFD, clientTransport.oSocket.fileno())
        self._lockManager.lockGlobal()
        try:
            monReport = self.__startReportToMonitoring()
        except Exception:
            monReport = False
        try:
            # Handshake
            try:
                result = clientTransport.handshake()
                if not result["OK"]:
                    clientTransport.close()
                    return
            except Exception:
                return
            # Add to the transport pool
            trid = self._transportPool.add(clientTransport)
            if not trid:
                return
            # Receive and check proposal
            result = self._receiveAndCheckProposal(trid)
            if not result["OK"]:
                self._transportPool.sendAndClose(trid, result)
                return
            proposalTuple = result["Value"]
            # Instantiate handler
            result = self._instantiateHandler(trid, proposalTuple)
            if not result["OK"]:
                self._transportPool.sendAndClose(trid, result)
                return
            handlerObj = result["Value"]
            # Execute the action
            result = self._processProposal(trid, proposalTuple, handlerObj)
            # Close the connection if required
            if result["closeTransport"] or not result["OK"]:
                if not result["OK"]:
                    gLogger.error("Error processing proposal", result["Message"])
                self._transportPool.close(trid)
            return result
        finally:
            self._lockManager.unlockGlobal()
            if monReport:
                self.__endReportToMonitoring(*monReport)

    def _createIdentityString(self, credDict, clientTransport=None):
        if "username" in credDict:
            if "group" in credDict:
                identity = "[%s:%s]" % (credDict["username"], credDict["group"])
            else:
                identity = "[%s:unknown]" % credDict["username"]
        else:
            identity = "unknown"
        if clientTransport:
            addr = clientTransport.getRemoteAddress()
            if addr:
                addr = "{%s:%s}" % (addr[0], addr[1])
        if "DN" in credDict:
            identity += "(%s)" % credDict["DN"]
        return identity

    @staticmethod
    def _deserializeProposalTuple(serializedProposal):
        """We receive the proposalTuple as a list.
        Turn it into a tuple again
        """
        proposalTuple = tuple(tuple(x) if isinstance(x, list) else x for x in serializedProposal)
        return proposalTuple

    def _receiveAndCheckProposal(self, trid):
        clientTransport = self._transportPool.get(trid)
        # Get the peer credentials
        credDict = clientTransport.getConnectingCredentials()
        # Receive the action proposal
        retVal = clientTransport.receiveData(1024)
        if not retVal["OK"]:
            gLogger.error(
                "Invalid action proposal",
                "%s %s" % (self._createIdentityString(credDict, clientTransport), retVal["Message"]),
            )
            return S_ERROR("Invalid action proposal")
        proposalTuple = Service._deserializeProposalTuple(retVal["Value"])
        gLogger.debug("Received action from client", "/".join(list(proposalTuple[1])))
        # Check if there are extra credentials
        if proposalTuple[2]:
            clientTransport.setExtraCredentials(proposalTuple[2])
        # Check if this is the requested service
        requestedService = proposalTuple[0][0]
        if requestedService not in self._validNames:
            return S_ERROR("%s is not up in this server" % requestedService)
        # Check if the action is valid
        requestedActionType = proposalTuple[1][0]
        if requestedActionType not in Service.SVC_VALID_ACTIONS:
            return S_ERROR("%s is not a known action type" % requestedActionType)
        # Check if it's authorized
        result = self._authorizeProposal(proposalTuple[1], trid, credDict)
        if not result["OK"]:
            return result
        # Proposal is OK
        return S_OK(proposalTuple)

    def _authorizeProposal(self, actionTuple, trid, credDict):
        # Find CS path for the Auth rules
        referedAction = self._isMetaAction(actionTuple[0])
        if referedAction:
            csAuthPath = "%s/Default" % actionTuple[0]
            hardcodedMethodAuth = self._actions["auth"][actionTuple[0]]
        else:
            if actionTuple[0] == "RPC":
                csAuthPath = actionTuple[1]
            else:
                csAuthPath = "/".join(actionTuple)
            # Find if there are hardcoded auth rules in the code
            hardcodedMethodAuth = False
            if actionTuple[0] in self._actions["auth"]:
                hardcodedRulesByType = self._actions["auth"][actionTuple[0]]
                if actionTuple[0] == "FileTransfer":
                    methodName = actionTuple[1][0].lower() + actionTuple[1][1:]
                else:
                    methodName = actionTuple[1]

                if methodName in hardcodedRulesByType:
                    hardcodedMethodAuth = hardcodedRulesByType[methodName]
        # Auth time!
        if not self._authMgr.authQuery(csAuthPath, credDict, hardcodedMethodAuth):
            # Get the identity string
            identity = self._createIdentityString(credDict)
            fromHost = "unknown host"
            tr = self._transportPool.get(trid)
            if tr:
                fromHost = "/".join([str(item) for item in tr.getRemoteAddress()])
            gLogger.warn(
                "Unauthorized query", "to %s:%s by %s from %s" % (self._name, "/".join(actionTuple), identity, fromHost)
            )
            result = S_ERROR(ENOAUTH, "Unauthorized query")
        else:
            result = S_OK()

        # Security log
        tr = self._transportPool.get(trid)
        if not tr:
            return S_ERROR("Client disconnected")
        sourceAddress = tr.getRemoteAddress()
        identity = self._createIdentityString(credDict)
        Service.SVC_SECLOG_CLIENT.addMessage(
            result["OK"],
            sourceAddress[0],
            sourceAddress[1],
            identity,
            self._cfg.getHostname(),
            self._cfg.getPort(),
            self._name,
            "/".join(actionTuple),
        )
        return result

    def _instantiateHandler(self, trid, proposalTuple=None):
        """
        Generate an instance of the handler for a given service

        :param int trid: transport ID
        :param tuple proposalTuple: tuple describing the proposed action

        :return: S_OK/S_ERROR, Value is the handler object
        """
        # Generate the client params
        clientParams = {"serviceStartTime": self._startTime}
        if proposalTuple:
            # The 4th element is the client version
            clientParams["clientVersion"] = proposalTuple[3] if len(proposalTuple) > 3 else None
            clientParams["clientSetup"] = proposalTuple[0][1]
            if len(proposalTuple[0]) < 3:
                clientParams["clientVO"] = gConfig.getValue("/DIRAC/VirtualOrganization", "unknown")
            else:
                clientParams["clientVO"] = proposalTuple[0][2]
        clientTransport = self._transportPool.get(trid)
        if clientTransport:
            clientParams["clientAddress"] = clientTransport.getRemoteAddress()
        # Generate handler dict with per client info
        handlerInitDict = dict(self._serviceInfoDict)
        for key in clientParams:
            handlerInitDict[key] = clientParams[key]
        # Instantiate and initialize
        try:
            handlerInstance = self._handler["class"](handlerInitDict, trid)
            handlerInstance.initialize()
        except Exception as e:
            gLogger.exception("Server error while loading handler: %s" % str(e))
            return S_ERROR("Server error while loading handler")
        return S_OK(handlerInstance)

    def _processProposal(self, trid, proposalTuple, handlerObj):
        # Notify the client we're ready to execute the action
        retVal = self._transportPool.send(trid, S_OK())
        if not retVal["OK"]:
            return retVal

        messageConnection = False
        if proposalTuple[1] == ("Connection", "new"):
            messageConnection = True

        if messageConnection:

            if self._msgBroker.getNumConnections() > self._cfg.getMaxMessagingConnections():
                result = S_ERROR("Maximum number of connections reached. Try later")
                result["closeTransport"] = True
                return result

            # This is a stable connection
            self._msgBroker.addTransportId(
                trid,
                self._name,
                receiveMessageCallback=self._mbReceivedMsg,
                disconnectCallback=self._mbDisconnect,
                listenToConnection=False,
            )

        result = self._executeAction(trid, proposalTuple, handlerObj)
        if result["OK"] and messageConnection:
            self._msgBroker.listenToTransport(trid)
            result = self._mbConnect(trid, handlerObj)
            if not result["OK"]:
                self._msgBroker.removeTransport(trid)

        result["closeTransport"] = not messageConnection or not result["OK"]
        return result

    def _mbConnect(self, trid, handlerObj=None):
        if not handlerObj:
            result = self._instantiateHandler(trid)
            if not result["OK"]:
                return result
            handlerObj = result["Value"]
        return handlerObj._rh_executeConnectionCallback("connected")

    def _executeAction(self, trid, proposalTuple, handlerObj):
        try:
            response = handlerObj._rh_executeAction(proposalTuple)
            if not response["OK"]:
                return response
            if self.activityMonitoring:
                self.activityMonitoringReporter.addRecord(
                    {
                        "timestamp": int(Time.toEpoch()),
                        "host": Network.getFQDN(),
                        "componentType": "service",
                        "component": "_".join(self._name.split("/")),
                        "componentLocation": self._cfg.getURL(),
                        "ServiceResponseTime": response["Value"][1],
                    }
                )
            return response["Value"][0]
        except Exception as e:
            gLogger.exception("Exception while executing handler action")
            return S_ERROR("Server error while executing action: %s" % str(e))

    def _mbReceivedMsg(self, trid, msgObj):
        result = self._authorizeProposal(
            ("Message", msgObj.getName()), trid, self._transportPool.get(trid).getConnectingCredentials()
        )
        if not result["OK"]:
            return result
        result = self._instantiateHandler(trid)
        if not result["OK"]:
            return result
        handlerObj = result["Value"]
        response = handlerObj._rh_executeMessageCallback(msgObj)
        if self.activityMonitoring and response["OK"]:
            self.activityMonitoringReporter.addRecord(
                {
                    "timestamp": int(Time.toEpoch()),
                    "host": Network.getFQDN(),
                    "componentType": "service",
                    "component": "_".join(self._name.split("/")),
                    "componentLocation": self._cfg.getURL(),
                    "ServiceResponseTime": response["Value"][1],
                }
            )
        if response["OK"]:
            return response["Value"][0]
        else:
            return response

    def _mbDisconnect(self, trid):
        result = self._instantiateHandler(trid)
        if not result["OK"]:
            return result
        handlerObj = result["Value"]
        return handlerObj._rh_executeConnectionCallback("drop")

    def __activityMonitoringReporting(self):
        """This method is called by the ThreadScheduler as a periodic task in order to commit the collected data which
        is done by the MonitoringReporter and is send to the 'ComponentMonitoring' type.

        :return: True / False
        """
        result = self.activityMonitoringReporter.commit()
        return result["OK"]

    def __startReportToMonitoring(self):
        if not self.activityMonitoring:
            self._monitor.addMark("Queries")

        now = time.time()
        stats = os.times()
        cpuTime = stats[0] + stats[2]
        if now - self.__monitorLastStatsUpdate < 0:
            return (now, cpuTime)
        # Send CPU consumption mark
        self.__monitorLastStatsUpdate = now
        # Send Memory consumption mark
        membytes = MemStat.VmB("VmRSS:")
        if membytes:
            mem = membytes / (1024.0 * 1024.0)
            if not self.activityMonitoring:
                self._monitor.addMark("MEM", mem)
        return (now, cpuTime)

    def __endReportToMonitoring(self, initialWallTime, initialCPUTime):
        wallTime = time.time() - initialWallTime
        stats = os.times()
        cpuTime = stats[0] + stats[2] - initialCPUTime
        percentage = cpuTime / wallTime * 100.0
        if percentage > 0:
            if not self.activityMonitoring:
                self._monitor.addMark("CPU", percentage)



""" DB is a base class for multiple DIRAC databases that are based on MySQL.
    It uniforms the way the database objects are constructed
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Base.DIRACDB import DIRACDB
from DIRAC.Core.Utilities.MySQL import MySQL
from DIRAC.ConfigurationSystem.Client.Utilities import getDBParameters

__RCSID__ = "$Id$"


class DB(DIRACDB, MySQL):
    """All DIRAC DB classes should inherit from this one (unless using sqlalchemy)"""

    def __init__(self, dbname, fullname, debug=False):

        self.fullname = fullname

        result = getDBParameters(fullname)
        if not result["OK"]:
            raise RuntimeError("Cannot get database parameters: %s" % result["Message"])

        dbParameters = result["Value"]
        self.dbHost = dbParameters["Host"]
        self.dbPort = dbParameters["Port"]
        self.dbUser = dbParameters["User"]
        self.dbPass = dbParameters["Password"]
        self.dbName = dbParameters["DBName"]

        super(DB, self).__init__(
            hostName=self.dbHost,
            userName=self.dbUser,
            passwd=self.dbPass,
            dbName=self.dbName,
            port=self.dbPort,
            debug=debug,
        )

        if not self._connected:
            raise RuntimeError("Can not connect to DB '%s', exiting..." % self.dbName)

        self.log.info("===================== MySQL ======================")
        self.log.info("User:           " + self.dbUser)
        self.log.info("Host:           " + self.dbHost)
        self.log.info("Port:           " + str(self.dbPort))
        # self.log.info("Password:       "+ self.dbPass)
        self.log.info("DBName:         " + self.dbName)
        self.log.info("==================================================")



"""
Test properties of log records
"""

__RCSID__ = "$Id$"

from six.moves import _thread as thread
import pytest

from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import gLogger, gLoggerReset
from DIRAC.FrameworkSystem.private.standardLogging.test.TestLogUtilities import cleaningLog, captureBackend


def test_showFormatOptionsInit():
    """
    Check showHeaders and showThreadIDs methods at initialization
    """
    capturedBackend, log, sublog = gLoggerReset()

    # first, make sure options are inherited from gLogger
    assert gLogger._options["headerIsShown"]
    assert gLogger._options["timeStampIsShown"]
    assert gLogger._options["contextIsShown"]
    assert not gLogger._options["threadIDIsShown"]
    assert log._options["headerIsShown"] == gLogger._options["headerIsShown"]
    assert log._options["timeStampIsShown"] == gLogger._options["timeStampIsShown"]
    assert log._options["contextIsShown"] == gLogger._options["contextIsShown"]
    assert log._options["threadIDIsShown"] == gLogger._options["threadIDIsShown"]
    assert sublog._options["headerIsShown"] == log._options["headerIsShown"]
    assert sublog._options["timeStampIsShown"] == log._options["timeStampIsShown"]
    assert sublog._options["contextIsShown"] == log._options["contextIsShown"]
    assert sublog._options["threadIDIsShown"] == log._options["threadIDIsShown"]

    # create log records and check that the format is correct
    gLogger.notice("me")
    logstring = cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    log.notice("ss")
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    sublog.notice("age")
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    assert logstring == "Framework NOTICE: me\nFramework/log NOTICE: ss\nFramework/log/sublog NOTICE: age\n"


@pytest.mark.parametrize(
    "header, threadID, timeStamp, context, msg,\
                          expectedLog, isThreadIDAvailable, isTimeStampAvailable",
    [
        (False, False, False, False, "msg", "msg\nmsg\nmsg\n", False, False),
        (True, False, False, False, "m", "NOTICE: m\nNOTICE: m\nNOTICE: m\n", False, False),
        (False, True, False, False, "msg", "msg\nmsg\nmsg\n", False, False),
        (True, True, False, False, "m", "[%s] NOTICE: m\n[%s] NOTICE: m\n[%s] NOTICE: m\n", True, False),
        (False, False, True, False, "msg", "msg\nmsg\nmsg\n", False, False),
        (True, False, True, False, "m", "NOTICE: m\nNOTICE: m\nNOTICE: m\n", False, True),
        (False, True, True, False, "msg", "msg\nmsg\nmsg\n", False, False),
        (True, True, True, False, "m", "[%s] NOTICE: m\n[%s] NOTICE: m\n[%s] NOTICE: m\n", True, True),
        (False, False, False, True, "msg", "msg\nmsg\nmsg\n", False, False),
        (
            True,
            False,
            False,
            True,
            "m",
            "Framework NOTICE: m\nFramework/log NOTICE: m\nFramework/log/sublog NOTICE: m\n",
            False,
            False,
        ),
        (False, True, False, True, "msg", "msg\nmsg\nmsg\n", False, False),
        (
            True,
            True,
            False,
            True,
            "m",
            "Framework [%s] NOTICE: m\nFramework/log [%s] NOTICE: m\n" + "Framework/log/sublog [%s] NOTICE: m\n",
            True,
            False,
        ),
        (False, False, True, True, "msg", "msg\nmsg\nmsg\n", False, False),
        (
            True,
            False,
            True,
            True,
            "m",
            "Framework NOTICE: m\nFramework/log NOTICE: m\nFramework/log/sublog NOTICE: m\n",
            False,
            True,
        ),
        (False, True, True, True, "msg", "msg\nmsg\nmsg\n", False, False),
        (
            True,
            True,
            True,
            True,
            "m",
            "Framework [%s] NOTICE: m\nFramework/log [%s] NOTICE: m\n" + "Framework/log/sublog [%s] NOTICE: m\n",
            True,
            True,
        ),
    ],
)
def test_showFormatOptionsgLogger(
    header, threadID, timeStamp, context, msg, expectedLog, isThreadIDAvailable, isTimeStampAvailable
):
    """
    Set gLogger options, check that options are inherited in log and sublog
    """
    capturedBackend, log, sublog = gLoggerReset()

    # setting these values should modify the way the log record is displayed
    gLogger.showHeaders(header)
    gLogger.showThreadIDs(threadID)
    gLogger.showTimeStamps(timeStamp)
    gLogger.showContexts(context)

    # log and sublog should inherit from the changes
    assert gLogger._options["headerIsShown"] == header
    assert gLogger._options["threadIDIsShown"] == threadID
    assert log._options["headerIsShown"] == gLogger._options["headerIsShown"]
    assert log._options["timeStampIsShown"] == gLogger._options["timeStampIsShown"]
    assert log._options["contextIsShown"] == gLogger._options["contextIsShown"]
    assert log._options["threadIDIsShown"] == gLogger._options["threadIDIsShown"]
    assert sublog._options["headerIsShown"] == log._options["headerIsShown"]
    assert sublog._options["timeStampIsShown"] == log._options["timeStampIsShown"]
    assert sublog._options["contextIsShown"] == log._options["contextIsShown"]
    assert sublog._options["threadIDIsShown"] == log._options["threadIDIsShown"]

    # create log records and check the format is correct
    gLogger.notice(msg)
    logValue = capturedBackend.getvalue()
    # check that timestamp is available if it has to be available
    assert ("UTC" in logValue) == isTimeStampAvailable
    logstring = cleaningLog(logValue)
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    log.notice(msg)
    logValue = capturedBackend.getvalue()
    assert ("UTC" in logValue) == isTimeStampAvailable
    logstring += cleaningLog(logValue)
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    sublog.notice(msg)
    logValue = capturedBackend.getvalue()
    assert ("UTC" in logValue) == isTimeStampAvailable
    logstring += cleaningLog(logValue)
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    # check that threadID is present in the log when it should be present
    threadIDValue = str(thread.get_ident())
    assert (threadIDValue in logstring) == isThreadIDAvailable
    # as thread ID depends on the execution, we have to add it to the expected results
    if isThreadIDAvailable:
        expectedLog = expectedLog % (threadIDValue, threadIDValue, threadIDValue)
    assert expectedLog == logstring


@pytest.mark.parametrize(
    "header, threadID, msg, expectedLog, isThreadIDAvailable",
    [
        (False, False, "message", "message\nmessage\n", False),
        (True, False, "message", "Framework/log NOTICE: message\nFramework/log/sublog NOTICE: message\n", False),
        (False, True, "message", "message\nmessage\n", False),
        (
            True,
            True,
            "message",
            "Framework/log [%s] NOTICE: message\nFramework/log/sublog [%s] NOTICE: message\n",
            True,
        ),
    ],
)
def test_showFormatOptionsLog(header, threadID, msg, expectedLog, isThreadIDAvailable):
    """
    Set log (child of gLogger) options, check that options are inherited in sublog
    """
    capturedBackend, log, sublog = gLoggerReset()

    # set gLogger options, they should not be modified
    gLogger.showHeaders(True)
    gLogger.showThreadIDs(False)

    # set log options
    log.showHeaders(header)
    log.showThreadIDs(threadID)

    # sublog should inherit from the changes, gLogger should not be affected
    assert gLogger._options["headerIsShown"]
    assert not gLogger._options["threadIDIsShown"]
    assert log._options["headerIsShown"] == header
    assert log._options["threadIDIsShown"] == threadID
    assert sublog._options["headerIsShown"] == log._options["headerIsShown"]
    assert sublog._options["threadIDIsShown"] == log._options["threadIDIsShown"]

    # create log records and check the format is correct
    gLogger.notice(msg)
    logstring = cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    log.notice(msg)
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    sublog.notice(msg)
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    expectedLog = "Framework NOTICE: message\n" + expectedLog

    threadIDValue = str(thread.get_ident())
    assert (threadIDValue in logstring) == isThreadIDAvailable
    # as thread ID depends on the execution, we have to add it to the expected results
    if isThreadIDAvailable:
        expectedLog = expectedLog % (threadIDValue, threadIDValue)
    assert expectedLog == logstring


@pytest.mark.parametrize(
    "header, threadID, msg, expectedLog, isThreadIDAvailable",
    [
        (False, False, "message", "message\n", False),
        (True, False, "message", "Framework/log/sublog NOTICE: message\n", False),
        (False, True, "message", "message\n", False),
        (True, True, "message", "Framework/log/sublog [%s] NOTICE: message\n", True),
    ],
)
def test_showFormatOptionsSubLog(header, threadID, msg, expectedLog, isThreadIDAvailable):
    """
    Set sublog (child of log) options
    """
    capturedBackend, log, sublog = gLoggerReset()

    # set gLogger and log options, sublog options should not be modified
    gLogger.showHeaders(True)
    gLogger.showThreadIDs(False)
    log.showHeaders(False)
    log.showThreadIDs(False)

    # set sublog options
    sublog.showHeaders(header)
    sublog.showThreadIDs(threadID)

    # log should inherit from the options of gLogger, subLog shoud not inherit from log
    assert gLogger._options["headerIsShown"]
    assert not gLogger._options["threadIDIsShown"]
    assert not log._options["headerIsShown"]
    assert not log._options["threadIDIsShown"]
    assert sublog._options["headerIsShown"] == header
    assert sublog._options["threadIDIsShown"] == threadID

    # create log records and check the format is correct
    gLogger.notice(msg)
    logstring = cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    log.notice(msg)
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    sublog.notice(msg)
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    expectedLog = "Framework NOTICE: message\nmessage\n" + expectedLog

    threadIDValue = str(thread.get_ident())
    assert (threadIDValue in logstring) == isThreadIDAvailable
    # as thread ID depends on the execution, we have to add it to the expected results
    if isThreadIDAvailable:
        expectedLog = expectedLog % threadIDValue
    assert expectedLog == logstring


def test_showFormatOptionsPropag():
    """
    Make sure log and sublog don't inherit from gLogger options if already set
    """
    capturedBackend, log, sublog = gLoggerReset()

    # set logger options
    log.showHeaders(False)
    log.showThreadIDs(False)

    # then, set gLogger options again
    gLogger.showHeaders(True)
    gLogger.showThreadIDs(True)

    # log should not inherit from the options of gLogger as it has been modified by a developer
    # subLog shoud inherit from log, it has not been modified
    assert gLogger._options["headerIsShown"]
    assert gLogger._options["threadIDIsShown"]
    assert log._options["headerIsShown"] != gLogger._options["headerIsShown"]
    assert log._options["threadIDIsShown"] != gLogger._options["threadIDIsShown"]
    assert sublog._options["headerIsShown"] == log._options["headerIsShown"]
    assert sublog._options["threadIDIsShown"] == log._options["threadIDIsShown"]

    # a log record is sent, we then get the result to see if options have been taken into account
    gLogger.notice("me")
    logstring = cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    log.notice("ss")
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    sublog.notice("age")
    logstring += cleaningLog(capturedBackend.getvalue())
    capturedBackend.truncate(0)
    capturedBackend.seek(0)

    threadID = str(thread.get_ident())
    expectedLog = "Framework [%s] NOTICE: me\nss\nage\n" % threadID

    assert expectedLog == logstring



########################################################################
# File :   PilotBundle.py
# Author : Ricardo Graciani
########################################################################
"""
  Collection of Utilities to handle pilot jobs

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id:$"

import os
import base64
import bz2
import tempfile


def bundleProxy(executableFile, proxy):
    """Create a self extracting archive bundling together an executable script and a proxy"""

    compressedAndEncodedProxy = base64.b64encode(bz2.compress(proxy.dumpAllToString()["Value"])).decode()
    with open(executableFile, "rb") as fp:
        compressedAndEncodedExecutable = base64.b64encode(bz2.compress(fp.read(), 9)).decode()

    bundle = """#!/usr/bin/env python
# Wrapper script for executable and proxy
import os
import tempfile
import sys
import stat
import base64
import bz2
import shutil

try:
  workingDirectory = tempfile.mkdtemp(suffix='_wrapper', prefix='TORQUE_')
  os.chdir(workingDirectory)
  open('proxy', "w").write(bz2.decompress(base64.b64decode("%(compressedAndEncodedProxy)s")))
  open('%(executable)s', "w").write(bz2.decompress(base64.b64decode("%(compressedAndEncodedExecutable)s")))
  os.chmod('proxy', stat.S_IRUSR | stat.S_IWUSR)
  os.chmod('%(executable)s', stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
  os.environ["X509_USER_PROXY"] = os.path.join(workingDirectory, 'proxy')
except Exception as x:
  print >> sys.stderr, x
  sys.exit(-1)
cmd = "./%(executable)s"
print 'Executing: ', cmd
sys.stdout.flush()
os.system(cmd)

shutil.rmtree(workingDirectory)

""" % {
        "compressedAndEncodedProxy": compressedAndEncodedProxy,
        "compressedAndEncodedExecutable": compressedAndEncodedExecutable,
        "executable": os.path.basename(executableFile),
    }

    return bundle


def writeScript(script, writeDir=None):
    """
    Write script into a temporary unique file under provided writeDir
    """
    fd, name = tempfile.mkstemp(suffix="_pilotWrapper.py", prefix="DIRAC_", dir=writeDir)
    pilotWrapper = os.fdopen(fd, "w")
    pilotWrapper.write(script)
    pilotWrapper.close()
    return name



""" For reporting messages to security log service
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import syslog

from DIRAC import gLogger, gConfig
from DIRAC.Core.DISET.RPCClient import RPCClient
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler


class SecurityLogClient(object):

    __securityLogStore = []

    def __init__(self):
        self.__messagesList = []
        self.__maxMessagesInBundle = 1000
        self.__maxMessagesWaiting = 10000
        self.__taskId = gThreadScheduler.addPeriodicTask(30, self.__sendData)

    def addMessage(
        self,
        success,
        sourceIP,
        sourcePort,
        sourceIdentity,
        destinationIP,
        destinationPort,
        destinationService,
        action,
        timestamp=False,
    ):
        if not timestamp:
            timestamp = Time.dateTime()
        msg = [
            timestamp,
            success,
            sourceIP,
            sourcePort,
            sourceIdentity,
            destinationIP,
            destinationPort,
            destinationService,
            action,
        ]
        if gConfig.getValue("/Registry/EnableSysLog", False):
            strMsg = "Time=%s Accept=%s Source=%s:%s SourceID=%s Destination=%s:%s Service=%s Action=%s"
            syslog.syslog(strMsg % msg)
        while len(self.__messagesList) > self.__maxMessagesWaiting:
            self.__messagesList.pop(0)
        if not self.__securityLogStore:
            self.__messagesList.append(msg)
        else:
            self.__securityLogStore[0].logAction(msg)

    def setLogStore(self, logStore):
        while self.__securityLogStore:
            self.__securityLogStore.pop()
        self.__securityLogStore.append(logStore)
        gThreadScheduler.addPeriodicTask(10, self.__sendData, executions=1)

    def __sendData(self):
        gLogger.debug("Sending records to security log service...")
        msgList = self.__messagesList
        self.__messagesList = []
        rpcClient = RPCClient("Framework/SecurityLogging")
        for _i in range(0, len(msgList), self.__maxMessagesInBundle):
            msgsToSend = msgList[: self.__maxMessagesInBundle]
            result = rpcClient.logActionBundle(msgsToSend)
            if not result["OK"]:
                self.__messagesList.extend(msgList)
                break
            msgList = msgList[self.__maxMessagesInBundle :]
        gLogger.debug("Data sent to security log service")



#!/usr/bin/env python
"""
Get the currently defined user data volume quotas

Usage:
  dirac-dms-user-quota [options]

Example:
  $ dirac-dms-user-quota
  Current quota found to be 0.0 GB
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=False)

    import DIRAC
    from DIRAC import gLogger, gConfig
    from DIRAC.Core.Security.ProxyInfo import getProxyInfo

    res = getProxyInfo(False, False)
    if not res["OK"]:
        gLogger.error("Failed to get client proxy information.", res["Message"])
        DIRAC.exit(2)
    proxyInfo = res["Value"]
    username = proxyInfo["username"]

    try:
        quota = gConfig.getValue("/Registry/DefaultStorageQuota", 0.0)
        quota = gConfig.getValue("/Registry/Users/%s/Quota" % username, quota)
        gLogger.notice("Current quota found to be %.1f GB" % quota)
        DIRAC.exit(0)
    except Exception as x:
        gLogger.exception("Failed to convert retrieved quota", "", x)
        DIRAC.exit(-1)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""
Uninstallation of a DIRAC component
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import socket

from DIRAC import exit as DIRACexit
from DIRAC import gLogger, S_OK
from DIRAC.Core.Utilities.PromptUser import promptUser
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.FrameworkSystem.Utilities import MonitoringUtilities
from DIRAC.FrameworkSystem.Client.ComponentMonitoringClient import ComponentMonitoringClient

__RCSID__ = "$Id$"

force = False


def setForce(opVal):
    global force
    force = True
    return S_OK()


@Script()
def main():
    global force

    from DIRAC.FrameworkSystem.Client.ComponentInstaller import gComponentInstaller

    gComponentInstaller.exitOnError = True

    Script.registerSwitch("f", "force", "Forces the removal of the logs", setForce)
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(
        (
            "System/Component: Full component name (ie: WorkloadManagement/Matcher)",
            "System:           Name of the DIRAC system (ie: WorkloadManagement)",
        )
    )
    Script.registerArgument(" Component:        Name of the DIRAC service (ie: Matcher)", mandatory=False)
    _, args = Script.parseCommandLine()

    if len(args) == 1:
        args = args[0].split("/")

    if len(args) < 2:
        Script.showHelp(exitCode=1)

    system = args[0]
    component = args[1]

    monitoringClient = ComponentMonitoringClient()
    result = monitoringClient.getInstallations(
        {"Instance": component, "UnInstallationTime": None}, {"System": system}, {"HostName": socket.getfqdn()}, True
    )
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    if len(result["Value"]) < 1:
        gLogger.warn("Given component does not exist")
        DIRACexit(1)
    if len(result["Value"]) > 1:
        gLogger.error("Too many components match")
        DIRACexit(1)

    removeLogs = False
    if force:
        removeLogs = True
    else:
        if result["Value"][0]["Component"]["Type"] in gComponentInstaller.componentTypes:
            result = promptUser("Remove logs?", ["y", "n"], "n")
            if result["OK"]:
                removeLogs = result["Value"] == "y"
            else:
                gLogger.error(result["Message"])
                DIRACexit(1)

    result = gComponentInstaller.uninstallComponent(system, component, removeLogs)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)

    result = MonitoringUtilities.monitorUninstallation(system, component)
    if not result["OK"]:
        gLogger.error(result["Message"])
        DIRACexit(1)
    gLogger.notice("Successfully uninstalled component %s/%s" % (system, component))
    DIRACexit()


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import functools
import six
from DIRAC.Core.Utilities.DIRACSingleton import DIRACSingleton


@six.add_metaclass(DIRACSingleton)
class ThreadConfig(threading.local):
    """This class allows to contain extra information when a call is done on behalf of
    somebody else. Typically, when a host performs the request on behalf of a user.
    It is not used inside DIRAC, but is used in WebAppDIRAC for example

    Note that the class is a singleton, meaning that you share the same object in the whole process,
    however the attributes are thread locals (because of the threading.local inheritance).

    Also, this class has to be populated manually, no Client class will do it for you.

    """

    def __init__(self):
        self.reset()

    def reset(self):
        """Reset extra information"""
        self.__DN = False
        self.__group = False
        self.__deco = False
        self.__setup = False

    def setDecorator(self, deco):
        """Set decorator

        :param deco: decorator
        """
        self.__deco = deco

    def getDecorator(self):
        """Return decorator

        :return: decorator
        """
        return self.__deco

    def setDN(self, DN):
        """Set DN

        :param str DN: DN
        """
        self.__DN = DN

    def getDN(self):
        """Return DN

        :return: str
        """
        return self.__DN

    def setGroup(self, group):
        """Set group

        :param str group: group name
        """
        self.__group = group

    def getGroup(self):
        """Return group name

        :return: str
        """
        return self.__group

    def setID(self, DN, group):
        """Set user ID

        :param str DN: user DN
        :param str group: user group
        """
        self.__DN = DN
        self.__group = group

    def getID(self):
        """Return user ID

        :return: tuple
        """
        return (self.__DN, self.__group)

    def setSetup(self, setup):
        """Set setup name

        :param str setup: setup name
        """
        self.__setup = setup

    def getSetup(self):
        """Return setup name

        :return: str
        """
        return self.__setup

    def dump(self):
        """Return extra information

        :return: tuple
        """
        return (self.__DN, self.__group, self.__setup)

    def load(self, tp):
        """Save extra information

        :param tuple tp: contain DN, group name, setup name
        """
        if tp[0]:
            self.__DN = tp[0]
        if tp[1]:
            self.__group = tp[1]
        if tp[2]:
            self.__setup = tp[2]


def threadDeco(method):
    """Tread decorator

    :param method: method

    :return: wrapped method
    """
    tc = ThreadConfig()

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        """Wrapper

        :return: wrapped method
        """
        deco = tc.getDecorator()
        if not deco:
            return method(*args, **kwargs)
        # Deco is a decorator sooo....
        return deco(method)(*args, **kwargs)

    return wrapper
